from copy import deepcopy
from datetime import datetime

from twisted.internet.defer import Deferred, succeed, inlineCallbacks
from twisted.trial.unittest import SkipTest, TestCase, FailTest

from vumi.blinkenlights.metrics import MetricMessage
from vumi.message import TransportUserMessage, TransportEvent
from vumi.tests.fake_amqp import FakeAMQPBroker, FakeAMQClient
from vumi.tests.helpers import (
    VumiTestCase, proxyable, generate_proxies, IHelper, import_skip,
    MessageHelper, WorkerHelper, MessageDispatchHelper, PersistenceHelper,
    success_result_of, RiakDisabledForTest)
from vumi.worker import BaseWorker


class TestHelperHelpers(TestCase):
    def test_proxyable(self):
        """
        @proxyable should set a `proxyable` attr on the func it decorates.
        """

        @proxyable
        def is_proxyable():
            pass
        self.assertTrue(hasattr(is_proxyable, 'proxyable'))

        def not_proxyable():
            pass
        self.assertFalse(hasattr(not_proxyable, 'proxyable'))

    def test_generate_proxies(self):
        """
        generate_proxies() should copy proxyable source attrs to target.
        """

        class Source(object):
            @proxyable
            def is_proxyable(self):
                return self

            def not_proxyable(self):
                pass

        class Target(object):
            pass

        source = Source()
        target = Target()

        self.assertFalse(hasattr(target, 'is_proxyable'))
        self.assertFalse(hasattr(target, 'not_proxyable'))

        generate_proxies(target, source)

        self.assertTrue(hasattr(target, 'is_proxyable'))
        self.assertFalse(hasattr(target, 'not_proxyable'))

        # `self` in both the original and proxied versions should be the source
        # rather than the target.
        self.assertEqual(source, source.is_proxyable())
        self.assertEqual(source, target.is_proxyable())

    def test_generate_proxies_multiple_sources(self):
        """
        generate_proxies() should copy attrs from multiple sources.
        """

        class Source1(object):
            @proxyable
            def is_proxyable_1(self):
                return self

        class Source2(object):
            @proxyable
            def is_proxyable_2(self):
                return self

        class Target(object):
            pass

        source1 = Source1()
        source2 = Source2()
        target = Target()

        self.assertFalse(hasattr(target, 'is_proxyable_1'))
        self.assertFalse(hasattr(target, 'is_proxyable_2'))

        generate_proxies(target, source1)
        generate_proxies(target, source2)

        self.assertTrue(hasattr(target, 'is_proxyable_1'))
        self.assertTrue(hasattr(target, 'is_proxyable_2'))

        # `self` in the proxied versions should be the appropriate source.
        self.assertEqual(source1, target.is_proxyable_1())
        self.assertEqual(source2, target.is_proxyable_2())

    def test_generate_proxies_multiple_sources_overlap(self):
        """
        generate_proxies() shouldn't copy proxyables with existing names.
        """

        class Source1(object):
            @proxyable
            def is_proxyable(self):
                return self

        class Source2(object):
            @proxyable
            def is_proxyable(self):
                return self

        class Target(object):
            pass

        source1 = Source1()
        source2 = Source2()
        target = Target()

        generate_proxies(target, source1)
        err = self.assertRaises(Exception, generate_proxies, target, source2)
        self.assertTrue('is_proxyable' in err.args[0])

    def test_import_skip_skips(self):
        """
        import_skip() should raise a SkipTest exception if given an ImportError
        referencing an expected module name.
        """
        try:
            import badmodule
            self.fail(
                "Expected ImportError for %r, nothing raised." % (badmodule,))
        except ImportError as import_error:
            self.assertRaises(SkipTest, import_skip, import_error, 'badmodule')

    def test_import_skip_reraises(self):
        """
        import_skip() should reraise the given ImportError if it does not
        reference an expected module name.
        """
        try:
            import badmodule
            self.fail(
                "Expected ImportError for %r, nothing raised." % (badmodule,))
        except ImportError as import_error:
            err = self.assertRaises(
                ImportError, import_skip, import_error, 'nothing')
            self.assertEqual(err, import_error)

    def test_success_result_of_no_result(self):
        """
        success_result_of() should raise a FailTest exception if the Deferred
        has no result.
        """
        d = Deferred()
        err = self.assertRaises(FailTest, success_result_of, d)
        self.assertEqual(
            err.args[0], "No result available for deferred: %r" % (d,))

    def test_success_result_of_failure(self):
        """
        success_result_of() should raise a FailTest exception if the Deferred
        has a failure result.
        """
        d = Deferred()
        d.errback(Exception())
        err = self.assertRaises(FailTest, success_result_of, d)
        self.assertTrue(err.args[0].startswith(
            "Expected success from deferred %r, got failure:" % (d,)))

    def test_success_result_of_success(self):
        """
        success_result_of() should raise a FailTest exception if the Deferred
        has no result.
        """
        d = Deferred()
        d.callback('foo')
        self.assertEqual(success_result_of(d), 'foo')


class TestMessageHelper(TestCase):
    def assert_message_fields(self, msg, field_dict):
        self.assertEqual(field_dict, dict(
            (k, v) for k, v in msg.payload.iteritems() if k in field_dict))

    def test_implements_IHelper(self):
        """
        MessageHelper instances should provide the IHelper interface.
        """
        self.assertTrue(IHelper.providedBy(MessageHelper()))

    def test_defaults(self):
        """
        MessageHelper instances should have the expected parameter defaults.
        """
        msg_helper = MessageHelper()
        self.assertEqual(msg_helper.transport_name, 'sphex')
        self.assertEqual(msg_helper.transport_type, 'sms')
        self.assertEqual(msg_helper.mobile_addr, '+41791234567')
        self.assertEqual(msg_helper.transport_addr, '9292')

    def test_setup_sync(self):
        """
        MessageHelper.setup() should return ``None``, not a Deferred.
        """
        msg_helper = MessageHelper()
        self.assertEqual(msg_helper.setup(), None)

    def test_cleanup_sync(self):
        """
        MessageHelper.cleanup() should return ``None``, not a Deferred.
        """
        msg_helper = MessageHelper()
        self.assertEqual(msg_helper.cleanup(), None)

    def test_make_inbound_defaults(self):
        """
        .make_inbound() should build a message with expected default values.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_inbound('inbound message')
        self.assert_message_fields(msg, {
            'content': 'inbound message',
            'from_addr': msg_helper.mobile_addr,
            'to_addr': msg_helper.transport_addr,
            'transport_type': msg_helper.transport_type,
            'transport_name': msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    def test_make_inbound_with_addresses(self):
        """
        .make_inbound() should use overridden addresses if provided.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_inbound(
            'inbound message', from_addr='ib_from', to_addr='ib_to')
        self.assert_message_fields(msg, {
            'content': 'inbound message',
            'from_addr': 'ib_from',
            'to_addr': 'ib_to',
            'transport_type': msg_helper.transport_type,
            'transport_name': msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    def test_make_inbound_with_helper_metadata(self):
        """
        .make_inbound() should use overridden helper_metadata if provided.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_inbound('inbound message', helper_metadata={
            'foo': {'bar': 'baz'},
            'quux': {},
        })
        self.assert_message_fields(msg, {
            'content': 'inbound message',
            'from_addr': msg_helper.mobile_addr,
            'to_addr': msg_helper.transport_addr,
            'transport_type': msg_helper.transport_type,
            'transport_name': msg_helper.transport_name,
            'helper_metadata': {
                'foo': {'bar': 'baz'},
                'quux': {},
            },
            'transport_metadata': {},
        })

    def test_make_outbound_defaults(self):
        """
        .make_outbound() should build a message with expected default values.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_outbound('outbound message')
        self.assert_message_fields(msg, {
            'content': 'outbound message',
            'from_addr': msg_helper.transport_addr,
            'to_addr': msg_helper.mobile_addr,
            'transport_type': msg_helper.transport_type,
            'transport_name': msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    def test_make_outbound_with_addresses(self):
        """
        .make_outbound() should use overridden addresses if provided.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_outbound(
            'outbound message', from_addr='ob_from', to_addr='ob_to')
        self.assert_message_fields(msg, {
            'content': 'outbound message',
            'from_addr': 'ob_from',
            'to_addr': 'ob_to',
            'transport_type': msg_helper.transport_type,
            'transport_name': msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    def test_make_outbound_with_helper_metadata(self):
        """
        .make_outbound() should use overridden helper_metadata if provided.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_outbound('outbound message', helper_metadata={
            'foo': {'bar': 'baz'},
            'quux': {},
        })
        self.assert_message_fields(msg, {
            'content': 'outbound message',
            'from_addr': msg_helper.transport_addr,
            'to_addr': msg_helper.mobile_addr,
            'transport_type': msg_helper.transport_type,
            'transport_name': msg_helper.transport_name,
            'helper_metadata': {
                'foo': {'bar': 'baz'},
                'quux': {},
            },
            'transport_metadata': {},
        })

    def test_make_user_message_defaults(self):
        """
        .make_user_message() should build a message with expected values.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_user_message('outbound message', 'from', 'to')
        expected_msg = TransportUserMessage(
            content='outbound message', from_addr='from', to_addr='to',
            transport_type=msg_helper.transport_type,
            transport_name=msg_helper.transport_name,
            transport_metadata={}, helper_metadata={},
            # These fields are generated in both messages, so copy them.
            message_id=msg['message_id'], timestamp=msg['timestamp'])
        self.assertEqual(expected_msg, msg)

    def test_make_user_message_all_fields(self):
        """
        .make_user_message() should build a message with all provided fields.
        """
        msg_helper = MessageHelper()
        msg_fields = {
            'content': 'outbound message',
            'from_addr': 'from',
            'from_addr_type': 'msisdn',
            'to_addr': 'to',
            'to_addr_type': 'mxit_id',
            'group': '#channel',
            'provider': 'MY-MNO',
            'session_event': TransportUserMessage.SESSION_NEW,
            'transport_type': 'irc',
            'transport_name': 'vuminet',
            'transport_metadata': {'foo': 'bar'},
            'helper_metadata': {'foo': {}},
            'in_reply_to': 'ccf9c2b9b1e94433be20d157e82786fe',
            'timestamp': datetime.utcnow(),
            'message_id': 'bbf9c2b9b1e94433be20d157e82786ed',
            'endpoint': 'foo_ep',
        }
        msg = msg_helper.make_user_message(**msg_fields)
        expected_fields = msg_fields.copy()
        expected_fields.update({
            'message_type': TransportUserMessage.MESSAGE_TYPE,
            'message_version': TransportUserMessage.MESSAGE_VERSION,
            'routing_metadata': {
                'endpoint_name': expected_fields.pop('endpoint'),
            }
        })
        self.assertEqual(expected_fields, msg.payload)

    def test_make_user_message_extra_fields(self):
        """
        .make_user_message() should build a message with extra fields.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_user_message(
            'outbound message', 'from', 'to', foo='bar', baz='quux')
        self.assert_message_fields(msg, {'foo': 'bar', 'baz': 'quux'})

    def test_make_event_defaults_ack(self):
        """
        .make_event() should build an ack event with expected values.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_event('ack', 'abc123', sent_message_id='sent')
        expected_event = TransportEvent(
            event_type='ack', user_message_id='abc123', sent_message_id='sent',
            transport_type=msg_helper.transport_type,
            transport_name=msg_helper.transport_name,
            transport_metadata={}, helper_metadata={},
            # These fields are generated in both messages, so copy them.
            event_id=event['event_id'], timestamp=event['timestamp'])
        self.assertEqual(expected_event, event)

    def test_make_event_defaults_nack(self):
        """
        .make_event() should build a nack event with expected values.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_event('nack', 'abc123', nack_reason='elves')
        expected_event = TransportEvent(
            event_type='nack', user_message_id='abc123', nack_reason='elves',
            transport_type=msg_helper.transport_type,
            transport_name=msg_helper.transport_name,
            transport_metadata={}, helper_metadata={},
            # These fields are generated in both messages, so copy them.
            event_id=event['event_id'], timestamp=event['timestamp'])
        self.assertEqual(expected_event, event)

    def test_make_event_defaults_dr(self):
        """
        .make_event() should build a delivery report with expected values.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_event(
            'delivery_report', 'abc123', delivery_status='pending')
        expected_event = TransportEvent(
            event_type='delivery_report', user_message_id='abc123',
            delivery_status='pending',
            transport_type=msg_helper.transport_type,
            transport_name=msg_helper.transport_name,
            transport_metadata={}, helper_metadata={},
            # These fields are generated in both messages, so copy them.
            event_id=event['event_id'], timestamp=event['timestamp'])
        self.assertEqual(expected_event, event)

    def test_make_event_all_fields(self):
        """
        .make_event() should build an event with all provided fields.
        """
        msg_helper = MessageHelper()
        event_fields = {
            'event_type': 'ack',
            'user_message_id': 'abc123',
            'sent_message_id': '123abc',
            'transport_type': 'irc',
            'transport_name': 'vuminet',
            'transport_metadata': {'foo': 'bar'},
            'helper_metadata': {'foo': {}},

            'timestamp': datetime.utcnow(),
            'event_id': 'e6b7efecda8e42988b1e6905ad40fae1',
            'endpoint': 'foo_ep',
        }
        event = msg_helper.make_event(**event_fields)
        expected_fields = event_fields.copy()
        expected_fields.update({
            'message_type': TransportEvent.MESSAGE_TYPE,
            'message_version': TransportEvent.MESSAGE_VERSION,
            'routing_metadata': {
                'endpoint_name': expected_fields.pop('endpoint'),
            }
        })
        self.assertEqual(expected_fields, event.payload)

    def test_make_event_extra_fields(self):
        """
        .make_event() should build an event with extra fields.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_event(
            'ack', 'abc123', sent_message_id='sent', foo='bar', baz='quux')
        self.assert_message_fields(event, {'foo': 'bar', 'baz': 'quux'})

    def test_make_ack_default(self):
        """
        .make_ack() should build an ack event with expected values.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_ack()
        self.assert_message_fields(event, {
            'event_type': 'ack',
            'sent_message_id': event['user_message_id'],
        })

    def test_make_ack_with_sent_message_id(self):
        """
        .make_ack() should build an ack with the provided sent_message_id.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_ack(sent_message_id='abc123')
        self.assert_message_fields(event, {
            'event_type': 'ack',
            'sent_message_id': 'abc123',
        })

    def test_make_ack_with_message(self):
        """
        .make_ack() should build an ack event for the provided message.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_outbound('test message')
        event = msg_helper.make_ack(msg)
        self.assert_message_fields(event, {
            'event_type': 'ack',
            'user_message_id': msg['message_id'],
            'sent_message_id': msg['message_id'],
        })

    def test_make_nack_default(self):
        """
        .make_nack() should build a nack event with expected values.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_nack()
        self.assert_message_fields(event, {
            'event_type': 'nack',
            'nack_reason': 'sunspots',
        })

    def test_make_nack_with_nack_reason(self):
        """
        .make_nack() should build a nack with the provided nack_reason.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_nack(nack_reason='bogon emissions')
        self.assert_message_fields(event, {
            'event_type': 'nack',
            'nack_reason': 'bogon emissions',
        })

    def test_make_nack_with_message(self):
        """
        .make_nack() should build a nack event for the provided message.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_outbound('test message')
        event = msg_helper.make_nack(msg)
        self.assert_message_fields(event, {
            'event_type': 'nack',
            'user_message_id': msg['message_id'],
            'nack_reason': 'sunspots',
        })

    def test_make_delivery_report_default(self):
        """
        .make_delivery_report() should build an event with expected values.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_delivery_report()
        self.assert_message_fields(event, {
            'event_type': 'delivery_report',
            'delivery_status': 'delivered',
        })

    def test_make_delivery_report_with_delivery_statuss(self):
        """
        .make_delivery_report() should build an event with the provided
        delivery_status.
        """
        msg_helper = MessageHelper()
        event = msg_helper.make_delivery_report(delivery_status='pending')
        self.assert_message_fields(event, {
            'event_type': 'delivery_report',
            'delivery_status': 'pending',
        })

    def test_make_delivery_report_with_message(self):
        """
        .make_delivery_report() should build an event for the provided message.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_outbound('test message')
        event = msg_helper.make_delivery_report(msg)
        self.assert_message_fields(event, {
            'event_type': 'delivery_report',
            'user_message_id': msg['message_id'],
            'delivery_status': 'delivered',
        })

    def test_make_reply(self):
        """
        .make_reply() should build a reply for the given message and content.
        """
        msg_helper = MessageHelper()
        msg = msg_helper.make_inbound('inbound')
        reply = msg_helper.make_reply(msg, 'reply content')
        self.assert_message_fields(reply, {
            'content': 'reply content',
            'to_addr': msg['from_addr'],
            'from_addr': msg['to_addr'],
            'in_reply_to': msg['message_id'],
        })


class FakeWorker(object):
    def __init__(self, stop_d=succeed(None)):
        self._stop_d = stop_d

    def stopWorker(self):
        return self._stop_d


class FakeBroker(object):
    def __init__(self, delivery_d=succeed(None)):
        self._delivery_d = delivery_d

    def wait_delivery(self):
        return self._delivery_d


class ToyWorker(BaseWorker):
    worker_started = False
    worker_stopped = False

    def setup_heartbeat(self):
        # Overriden to skip heartbeat setup.
        pass

    def setup_connectors(self):
        pass

    def setup_worker(self):
        self.worker_started = True

    def teardown_worker(self):
        self.worker_stopped = True


class TestWorkerHelper(VumiTestCase):
    def test_implements_IHelper(self):
        """
        WorkerHelper instances should provide the IHelper interface.
        """
        self.assertTrue(IHelper.providedBy(WorkerHelper()))

    def test_defaults(self):
        """
        WorkerHelper instances should have the expected parameter defaults.
        """
        worker_helper = WorkerHelper()
        self.assertEqual(worker_helper._connector_name, None)
        self.assertIsInstance(worker_helper.broker, FakeAMQPBroker)

    def test_all_params(self):
        """
        WorkerHelper should use the provided broker and connector name.
        """
        broker = FakeBroker()
        worker_helper = WorkerHelper("my_connector", broker)
        self.assertEqual(worker_helper._connector_name, "my_connector")
        self.assertEqual(worker_helper.broker, broker)

    def test_setup_sync(self):
        """
        WorkerHelper.setup() should return ``None``, not a Deferred.
        """
        worker_helper = WorkerHelper()
        self.assertEqual(worker_helper.setup(), None)

    def test_cleanup(self):
        """
        WorkerHelper.cleanup() should wait for broker delivery and stop all
        workers.
        """
        delivery_d = Deferred()
        worker_stop_d = Deferred()
        worker_helper = WorkerHelper(broker=FakeBroker(delivery_d=delivery_d))
        worker_helper._workers.append(FakeWorker(stop_d=worker_stop_d))
        d = worker_helper.cleanup()
        self.assertFalse(d.called)
        delivery_d.callback(None)
        self.assertFalse(d.called)
        worker_stop_d.callback(None)
        self.assertTrue(d.called)

    def test_cleanup_worker(self):
        """
        WorkerHelper.cleanup_worker() should remove the worker from its list
        and then stop it.
        """
        worker_stop_d = Deferred()
        worker = FakeWorker(stop_d=worker_stop_d)
        worker_helper = WorkerHelper()
        worker_helper._workers.append(worker)
        d = worker_helper.cleanup_worker(worker)
        self.assertEqual(worker_helper._workers, [])
        self.assertEqual(d, worker_stop_d)

    def test_get_fake_amqp_client(self):
        """
        WorkerHelper.get_fake_amqp_client() should return a FakeAMQClient
        wrapping the given broker.
        """
        broker = FakeBroker()
        client = WorkerHelper.get_fake_amqp_client(broker)
        self.assertIsInstance(client, FakeAMQClient)
        self.assertEqual(client.broker, broker)

    def test_get_worker_raw(self):
        """
        WorkerHelper.get_worker_raw() should create an instance of the given
        worker class with the given config.
        """
        broker = FakeBroker()
        worker = WorkerHelper.get_worker_raw(
            BaseWorker, {'foo': 'bar'}, broker)
        self.assertIsInstance(worker, BaseWorker)
        self.assertIsInstance(worker._amqp_client, FakeAMQClient)
        self.assertEqual(worker._amqp_client.broker, broker)
        self.assertEqual(worker.config, {
            'foo': 'bar',
            # worker_name is added for us if we don't provide one.
            'worker_name': 'unnamed',
        })

    def test_get_worker_raw_worker_name_configured(self):
        """
        WorkerHelper.get_worker_raw() should not overwrite worker_name.
        """
        broker = FakeBroker()
        worker = WorkerHelper.get_worker_raw(
            BaseWorker, {'worker_name': 'Gilbert'}, broker)
        self.assertIsInstance(worker, BaseWorker)
        self.assertIsInstance(worker._amqp_client, FakeAMQClient)
        self.assertEqual(worker._amqp_client.broker, broker)
        self.assertEqual(worker.config, {'worker_name': 'Gilbert'})

    def test_get_worker_raw_config_None(self):
        """
        WorkerHelper.get_worker_raw() can take ``None`` as a config.
        """
        broker = FakeBroker()
        worker = WorkerHelper.get_worker_raw(
            BaseWorker, None, broker)
        self.assertIsInstance(worker, BaseWorker)
        self.assertIsInstance(worker._amqp_client, FakeAMQClient)
        self.assertEqual(worker._amqp_client.broker, broker)
        self.assertEqual(worker.config, {})

    def test_get_worker(self):
        """
        WorkerHelper.get_worker() should create an instance of the given worker
        class with the given config, start it and return it.
        """
        worker_helper = WorkerHelper()
        worker_d = worker_helper.get_worker(ToyWorker, {'foo': 'bar'})
        worker = success_result_of(worker_d)
        self.assertIsInstance(worker, ToyWorker)
        self.assertIsInstance(worker._amqp_client, FakeAMQClient)
        self.assertEqual(worker._amqp_client.broker, worker_helper.broker)
        self.assertEqual(worker.config, {
            'foo': 'bar',
            # worker_name is added for us if we don't provide one.
            'worker_name': 'unnamed',
        })
        self.assertTrue(worker.worker_started)

    def test_get_worker_no_start(self):
        """
        WorkerHelper.get_worker() should not start the worker if asked not to.
        """
        worker_helper = WorkerHelper()
        worker_d = worker_helper.get_worker(ToyWorker, {}, start=False)
        worker = success_result_of(worker_d)
        self.assertIsInstance(worker, ToyWorker)
        self.assertIsInstance(worker._amqp_client, FakeAMQClient)
        self.assertEqual(worker._amqp_client.broker, worker_helper.broker)
        self.assertFalse(worker.worker_started)

    def _add_to_dispatched(self, broker, rkey, msg, kick=False):
        broker.exchange_declare('vumi', 'direct')
        broker.publish_message('vumi', rkey, msg)
        if kick:
            return broker.kick_delivery()

    def _add_to_dispatched_metrics(self, broker, msg):
        broker.exchange_declare('vumi.metrics', 'direct')
        broker.publish_message('vumi.metrics', 'vumi.metrics', msg)

    def test_get_dispatched(self):
        """
        WorkerHelper.get_dispatched() should get messages dispatched by the
        broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        dispatched = worker_helper.get_dispatched(
            'fooconn', 'inbound', TransportUserMessage)
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg)
        dispatched = worker_helper.get_dispatched(
            'fooconn', 'inbound', TransportUserMessage)
        self.assertEqual(dispatched, [msg])

    def test_get_dispatched_None_connector(self):
        """
        WorkerHelper.get_dispatched() should use the default connector if
        `None` is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        dispatched = worker_helper.get_dispatched(
            None, 'inbound', TransportUserMessage)
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg)
        dispatched = worker_helper.get_dispatched(
            None, 'inbound', TransportUserMessage)
        self.assertEqual(dispatched, [msg])

    def test_clear_all_dispatched(self):
        """
        WorkerHelper.clear_all_dispatched() should clear all messages
        dispatched by the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(broker, 'fooconn.inbound', msg)
        self._add_to_dispatched_metrics(broker, MetricMessage())
        self.assertNotEqual(broker.dispatched['vumi'], {})
        self.assertNotEqual(broker.dispatched['vumi.metrics'], {})
        worker_helper.clear_all_dispatched()
        self.assertEqual(broker.dispatched['vumi'], {})
        self.assertEqual(broker.dispatched['vumi.metrics'], {})

    def test_clear_all_dispatched_no_messages(self):
        """
        WorkerHelper.clear_all_dispatched() should succeed even if no messages
        have ever been dispatched.
        """
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        self.assertEqual(broker.dispatched, {})
        worker_helper.clear_all_dispatched()
        self.assertEqual(broker.dispatched, {})

    def test_get_dispatched_events(self):
        """
        WorkerHelper.get_dispatched_events() should get events dispatched by
        the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        dispatched = worker_helper.get_dispatched_events('fooconn')
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_ack()
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.event', msg)
        dispatched = worker_helper.get_dispatched_events('fooconn')
        self.assertEqual(dispatched, [msg])

    def test_get_dispatched_events_no_connector(self):
        """
        WorkerHelper.get_dispatched_events() should use the default connector
        if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        dispatched = worker_helper.get_dispatched_events()
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_ack()
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.event', msg)
        dispatched = worker_helper.get_dispatched_events()
        self.assertEqual(dispatched, [msg])

    def test_get_dispatched_inbound(self):
        """
        WorkerHelper.get_dispatched_inbound() should get inbound messages
        dispatched by the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        dispatched = worker_helper.get_dispatched_inbound('fooconn')
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg)
        dispatched = worker_helper.get_dispatched_inbound('fooconn')
        self.assertEqual(dispatched, [msg])

    def test_get_dispatched_inbound_no_connector(self):
        """
        WorkerHelper.get_dispatched_inbound() should use the default connector
        if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        dispatched = worker_helper.get_dispatched_inbound()
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg)
        dispatched = worker_helper.get_dispatched_inbound()
        self.assertEqual(dispatched, [msg])

    def test_get_dispatched_outbound(self):
        """
        WorkerHelper.get_dispatched_outbound() should get outbound messages
        dispatched by the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        dispatched = worker_helper.get_dispatched_outbound('fooconn')
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_outbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.outbound', msg)
        dispatched = worker_helper.get_dispatched_outbound('fooconn')
        self.assertEqual(dispatched, [msg])

    def test_get_dispatched_outbound_no_connector(self):
        """
        WorkerHelper.get_dispatched_outbound() should use the default connector
        if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        dispatched = worker_helper.get_dispatched_outbound()
        self.assertEqual(dispatched, [])
        msg = msg_helper.make_outbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.outbound', msg)
        dispatched = worker_helper.get_dispatched_outbound()
        self.assertEqual(dispatched, [msg])

    @inlineCallbacks
    def test_wait_for_dispatched_events(self):
        """
        WorkerHelper.wait_for_dispatched_events() should wait for events
        dispatched by the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        d = worker_helper.wait_for_dispatched_events(1, 'fooconn')
        self.assertEqual(d.called, False)
        msg = msg_helper.make_ack()
        yield self._add_to_dispatched(
            worker_helper.broker, 'fooconn.event', msg, kick=True)
        dispatched = success_result_of(d)
        self.assertEqual(dispatched, [msg])

    @inlineCallbacks
    def test_wait_for_dispatched_events_no_connector(self):
        """
        WorkerHelper.wait_for_dispatched_events() should get use the default
        connector if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        d = worker_helper.wait_for_dispatched_events(1)
        self.assertEqual(d.called, False)
        msg = msg_helper.make_ack()
        yield self._add_to_dispatched(
            worker_helper.broker, 'fooconn.event', msg, kick=True)
        dispatched = success_result_of(d)
        self.assertEqual(dispatched, [msg])

    @inlineCallbacks
    def test_wait_for_dispatched_inbound(self):
        """
        WorkerHelper.wait_for_dispatched_inbound() should wait for
        inbound messages dispatched by the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        d = worker_helper.wait_for_dispatched_inbound(1, 'fooconn')
        msg = msg_helper.make_inbound('message')
        yield self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg, kick=True)
        dispatched = success_result_of(d)
        self.assertEqual(dispatched, [msg])

    @inlineCallbacks
    def test_wait_for_dispatched_inbound_no_connector(self):
        """
        WorkerHelper.wait_for_dispatched_inbound() should use the default
        connector if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        d = worker_helper.wait_for_dispatched_inbound(1)
        msg = msg_helper.make_inbound('message')
        yield self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg, kick=True)
        dispatched = success_result_of(d)
        self.assertEqual(dispatched, [msg])

    @inlineCallbacks
    def test_wait_for_dispatched_outbound(self):
        """
        WorkerHelper.wait_for_dispatched_outbound() should wait for outbound
        messages dispatched by the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        d = worker_helper.wait_for_dispatched_outbound(1, 'fooconn')
        msg = msg_helper.make_outbound('message')
        yield self._add_to_dispatched(
            worker_helper.broker, 'fooconn.outbound', msg, kick=True)
        dispatched = success_result_of(d)
        self.assertEqual(dispatched, [msg])

    @inlineCallbacks
    def test_wait_for_dispatched_outbound_no_connector(self):
        """
        WorkerHelper.wait_for_dispatched_outbound() should use the default
        connector if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        d = worker_helper.wait_for_dispatched_outbound(1)
        msg = msg_helper.make_outbound('message')
        yield self._add_to_dispatched(
            worker_helper.broker, 'fooconn.outbound', msg, kick=True)
        dispatched = success_result_of(d)
        self.assertEqual(dispatched, [msg])

    def test_clear_dispatched_events(self):
        """
        WorkerHelper.clear_dispatched_events() should clear events messages
        dispatched to a particular endpoint from the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        msg = msg_helper.make_ack()
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.event', msg)
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.event'], [])
        worker_helper.clear_dispatched_events('fooconn')
        self.assertEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.event'], [])

    def test_clear_dispatched_events_no_connector(self):
        """
        WorkerHelper.clear_dispatched_events() should use the default connector
        if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        msg = msg_helper.make_ack()
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.event', msg)
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.event'], [])
        worker_helper.clear_dispatched_events()
        self.assertEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.event'], [])

    def test_clear_dispatched_inbound(self):
        """
        WorkerHelper.clear_dispatched_inbound() should clear inbound messages
        dispatched to a particular endpoint from the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg)
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.inbound'], [])
        worker_helper.clear_dispatched_inbound('fooconn')
        self.assertEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.inbound'], [])

    def test_clear_dispatched_inbound_no_connector(self):
        """
        WorkerHelper.clear_dispatched_inbound() should use the default
        connector if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        msg = msg_helper.make_inbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.inbound', msg)
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.inbound'], [])
        worker_helper.clear_dispatched_inbound()
        self.assertEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.inbound'], [])

    def test_clear_dispatched_outbound(self):
        """
        WorkerHelper.clear_dispatched_outbound() should clear outbound messages
        dispatched to a particular endpoint from the broker.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        msg = msg_helper.make_outbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.outbound', msg)
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.outbound'], [])
        worker_helper.clear_dispatched_outbound('fooconn')
        self.assertEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.outbound'], [])

    def test_clear_dispatched_outbound_no_connector(self):
        """
        WorkerHelper.clear_dispatched_outbound() should use the default
        connector if none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        msg = msg_helper.make_outbound('message')
        self._add_to_dispatched(
            worker_helper.broker, 'fooconn.outbound', msg)
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.outbound'], [])
        worker_helper.clear_dispatched_outbound()
        self.assertEqual(
            worker_helper.broker.dispatched['vumi']['fooconn.outbound'], [])

    @inlineCallbacks
    def test_dispatch_raw(self):
        """
        WorkerHelper.dispatch_raw() should dispatch a message.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.foo'), [])
        msg = msg_helper.make_inbound('message')
        yield worker_helper.dispatch_raw('fooconn.foo', msg)
        self.assertEqual(broker.get_messages('vumi', 'fooconn.foo'), [msg])

    @inlineCallbacks
    def test_dispatch_raw_with_exchange(self):
        """
        WorkerHelper.dispatch_raw() should dispatch a message on the specified
        exchange.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        broker.exchange_declare('blah', 'direct')
        self.assertEqual(broker.get_messages('blah', 'fooconn.foo'), [])
        msg = msg_helper.make_inbound('message')
        yield worker_helper.dispatch_raw('fooconn.foo', msg, exchange='blah')
        self.assertEqual(broker.get_messages('blah', 'fooconn.foo'), [msg])

    @inlineCallbacks
    def test_dispatch_event(self):
        """
        WorkerHelper.dispatch_event() should dispatch an event message.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        msg = msg_helper.make_ack()
        yield worker_helper.dispatch_event(msg, 'fooconn')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [msg])

    @inlineCallbacks
    def test_dispatch_event_no_connector(self):
        """
        WorkerHelper.dispatch_event() should use the default connector if
        none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        msg = msg_helper.make_ack()
        yield worker_helper.dispatch_event(msg)
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [msg])

    @inlineCallbacks
    def test_dispatch_inbound(self):
        """
        WorkerHelper.dispatch_inbound() should dispatch an inbound message.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [])
        msg = msg_helper.make_inbound('message')
        yield worker_helper.dispatch_inbound(msg, 'fooconn')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [msg])

    @inlineCallbacks
    def test_dispatch_inbound_no_connector(self):
        """
        WorkerHelper.dispatch_inbound() should use the default connector if
        none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [])
        msg = msg_helper.make_inbound('message')
        yield worker_helper.dispatch_inbound(msg)
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [msg])

    @inlineCallbacks
    def test_dispatch_outbound(self):
        """
        WorkerHelper.dispatch_outbound() should dispatch an outbound message.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper()
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.outbound'), [])
        msg = msg_helper.make_outbound('message')
        yield worker_helper.dispatch_outbound(msg, 'fooconn')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.outbound'), [msg])

    @inlineCallbacks
    def test_dispatch_outbound_no_connector(self):
        """
        WorkerHelper.dispatch_outbound() should use the default connector if
        none is passed in.
        """
        msg_helper = MessageHelper()
        worker_helper = WorkerHelper(connector_name='fooconn')
        broker = worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.outbound'), [])
        msg = msg_helper.make_outbound('message')
        yield worker_helper.dispatch_outbound(msg)
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.outbound'), [msg])

    def test_get_dispatched_metrics(self):
        """
        WorkerHelper.get_dispatched_metrics() should get metrics dispatched by
        the broker.
        """
        worker_helper = WorkerHelper()
        dispatched = worker_helper.get_dispatched_metrics()
        self.assertEqual(dispatched, [])

        msg = MetricMessage()
        self._add_to_dispatched_metrics(worker_helper.broker, msg)
        dispatched = worker_helper.get_dispatched_metrics()
        self.assertEqual(dispatched, [[]])

        msg = MetricMessage()
        msg.append('fake metric 1')
        msg.append('fake metric 2')
        self._add_to_dispatched_metrics(worker_helper.broker, msg)
        dispatched = worker_helper.get_dispatched_metrics()
        self.assertEqual(dispatched, [[], ['fake metric 1', 'fake metric 2']])

    def test_clear_dispatched_metrics(self):
        """
        WorkerHelper.clear_dispatched_metrics() should clear metrics messages
        dispatched from the broker.
        """
        worker_helper = WorkerHelper()
        self._add_to_dispatched_metrics(worker_helper.broker, MetricMessage())
        self.assertNotEqual(
            worker_helper.broker.dispatched['vumi.metrics'], {})
        worker_helper.clear_dispatched_metrics()
        self.assertEqual(
            worker_helper.broker.dispatched['vumi.metrics'], {})


class TestMessageDispatchHelper(VumiTestCase):
    def assert_message_fields(self, msg, field_dict):
        self.assertEqual(field_dict, dict(
            (k, v) for k, v in msg.payload.iteritems() if k in field_dict))

    def test_implements_IHelper(self):
        """
        MessageDispatchHelper instances should provide the IHelper interface.
        """
        self.assertTrue(IHelper.providedBy(MessageDispatchHelper(None, None)))

    def test_setup_sync(self):
        """
        MessageDispatchHelper.setup() should return ``None``, not a Deferred.
        """
        md_helper = MessageDispatchHelper(None, None)
        self.assertEqual(md_helper.setup(), None)

    def test_cleanup_sync(self):
        """
        MessageDispatchHelper.cleanup() should return ``None``, not a Deferred.
        """
        md_helper = MessageDispatchHelper(None, None)
        self.assertEqual(md_helper.cleanup(), None)

    @inlineCallbacks
    def test_make_dispatch_inbound_defaults(self):
        """
        .make_dispatch_inbound() should build and dispatch an inbound message.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [])
        msg = yield md_helper.make_dispatch_inbound('inbound message')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [msg])
        self.assert_message_fields(msg, {
            'content': 'inbound message',
            'from_addr': md_helper.msg_helper.mobile_addr,
            'to_addr': md_helper.msg_helper.transport_addr,
            'transport_type': md_helper.msg_helper.transport_type,
            'transport_name': md_helper.msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    @inlineCallbacks
    def test_make_dispatch_inbound_with_addresses(self):
        """
        .make_dispatch_inbound() should build and dispatch an inbound message
        with non-default parameters.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [])
        msg = yield md_helper.make_dispatch_inbound(
            'inbound message', from_addr='ib_from', to_addr='ib_to')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.inbound'), [msg])
        self.assert_message_fields(msg, {
            'content': 'inbound message',
            'from_addr': 'ib_from',
            'to_addr': 'ib_to',
            'transport_type': md_helper.msg_helper.transport_type,
            'transport_name': md_helper.msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    @inlineCallbacks
    def test_make_dispatch_outbound_defaults(self):
        """
        .make_dispatch_outbound() should build and dispatch an outbound
        message.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.outbound'), [])
        msg = yield md_helper.make_dispatch_outbound('outbound message')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.outbound'), [msg])
        self.assert_message_fields(msg, {
            'content': 'outbound message',
            'from_addr': md_helper.msg_helper.transport_addr,
            'to_addr': md_helper.msg_helper.mobile_addr,
            'transport_type': md_helper.msg_helper.transport_type,
            'transport_name': md_helper.msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    @inlineCallbacks
    def test_make_dispatch_outbound_with_addresses(self):
        """
        .make_dispatch_outbound() should build and dispatch an outbound message
        with non-default parameters.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.outbound'), [])
        msg = yield md_helper.make_dispatch_outbound(
            'outbound message', from_addr='ob_from', to_addr='ob_to')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.outbound'), [msg])
        self.assert_message_fields(msg, {
            'content': 'outbound message',
            'from_addr': 'ob_from',
            'to_addr': 'ob_to',
            'transport_type': md_helper.msg_helper.transport_type,
            'transport_name': md_helper.msg_helper.transport_name,
            'helper_metadata': {},
            'transport_metadata': {},
        })

    @inlineCallbacks
    def test_make_dispatch_ack_default(self):
        """
        .make_dispatch_ack() should build and dispatch an ack event.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        event = yield md_helper.make_dispatch_ack()
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.event'), [event])
        self.assert_message_fields(event, {
            'event_type': 'ack',
            'sent_message_id': event['user_message_id'],
        })

    @inlineCallbacks
    def test_make_ack_with_sent_message_id(self):
        """
        .make_dispatch_ack() should build and dispatch an ack event with
        non-default parameters.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        msg = md_helper.msg_helper.make_outbound('test message')
        event = yield md_helper.make_dispatch_ack(
            msg, sent_message_id='abc123')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.event'), [event])
        self.assert_message_fields(event, {
            'event_type': 'ack',
            'user_message_id': msg['message_id'],
            'sent_message_id': 'abc123',
        })

    @inlineCallbacks
    def test_make_dispatch_nack_default(self):
        """
        .make_dispatch_nack() should build and dispatch a nack event.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        event = yield md_helper.make_dispatch_nack()
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.event'), [event])
        self.assert_message_fields(event, {
            'event_type': 'nack',
            'nack_reason': 'sunspots',
        })

    @inlineCallbacks
    def test_make_nack_with_sent_message_id(self):
        """
        .make_dispatch_nack() should build and dispatch a nack event with
        non-default parameters.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        msg = md_helper.msg_helper.make_outbound('test message')
        event = yield md_helper.make_dispatch_nack(
            msg, nack_reason='bogon emissions')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.event'), [event])
        self.assert_message_fields(event, {
            'event_type': 'nack',
            'user_message_id': msg['message_id'],
            'nack_reason': 'bogon emissions',
        })

    @inlineCallbacks
    def test_make_dispatch_delivery_report_default(self):
        """
        .make_dispatch_delivery_report() should build and dispatch a
        delivery_report.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        event = yield md_helper.make_dispatch_delivery_report()
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.event'), [event])
        self.assert_message_fields(event, {
            'event_type': 'delivery_report',
            'delivery_status': 'delivered',
        })

    @inlineCallbacks
    def test_make_delivery_report_with_sent_message_id(self):
        """
        .make_dispatch_delivery_report() should build and dispatch a delivery
        report with non-default parameters.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.event'), [])
        msg = md_helper.msg_helper.make_outbound('test message')
        event = yield md_helper.make_dispatch_delivery_report(
            msg, delivery_status='pending')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.event'), [event])
        self.assert_message_fields(event, {
            'event_type': 'delivery_report',
            'user_message_id': msg['message_id'],
            'delivery_status': 'pending',
        })

    @inlineCallbacks
    def test_make_reply(self):
        """
        .make_dispatch_reply() should build and dispatch a reply message.
        """
        md_helper = MessageDispatchHelper(
            MessageHelper(), WorkerHelper('fooconn'))
        broker = md_helper.worker_helper.broker
        broker.exchange_declare('vumi', 'direct')
        self.assertEqual(broker.get_messages('vumi', 'fooconn.outbound'), [])
        msg = md_helper.msg_helper.make_inbound('inbound')
        reply = yield md_helper.make_dispatch_reply(msg, 'reply content')
        self.assertEqual(
            broker.get_messages('vumi', 'fooconn.outbound'), [reply])
        self.assert_message_fields(reply, {
            'content': 'reply content',
            'to_addr': msg['from_addr'],
            'from_addr': msg['to_addr'],
            'in_reply_to': msg['message_id'],
        })


class FakeRiakManagerForCleanup(object):
    purged = False

    def __init__(self, bucket_prefix, conns=None):
        self.bucket_prefix = bucket_prefix
        self.fake_conns = conns or []

    def purge_all(self):
        self.purged = True
        return 'maybe async'

    def close_manager(self):
        for conn in self.fake_conns:
            conn.close()
        self.fake_conns = []


class FakeRiakClientConnection(object):
    closed = False

    def close(self):
        self.closed = True


class FakeRedisManagerForCleanup(object):
    purged = False
    connected = True

    def __init__(self, key_prefix):
        self._key_prefix = key_prefix

    def _purge_all(self):
        if not self.connected:
            raise RuntimeError('Not connected')
        self.purged = True

    def close_manager(self):
        self.connected = False


class TestPersistenceHelper(VumiTestCase):
    @property
    def _RiakManager(self):
        try:
            from vumi.persist.riak_manager import RiakManager
        except ImportError, e:
            import_skip(e, 'riak')
        return RiakManager

    @property
    def _TxRiakManager(self):
        try:
            from vumi.persist.txriak_manager import TxRiakManager
        except ImportError, e:
            import_skip(e, 'riak')
        return TxRiakManager

    @property
    def _RedisManager(self):
        try:
            from vumi.persist.redis_manager import RedisManager
        except ImportError, e:
            import_skip(e, 'redis')
        return RedisManager

    @property
    def _TxRedisManager(self):
        from vumi.persist.txredis_manager import TxRedisManager
        return TxRedisManager

    def test_implements_IHelper(self):
        """
        PersistenceHelper instances should provide the IHelper interface.
        """
        self.assertTrue(IHelper.providedBy(PersistenceHelper()))

    def test_defaults(self):
        """
        PersistenceHelper instances should have the expected parameter
        defaults.
        """
        persistence_helper = PersistenceHelper()
        self.assertEqual(persistence_helper.use_riak, False)
        self.assertEqual(persistence_helper.is_sync, False)

    def test_all_params(self):
        """
        PersistenceHelper instances should accept ``use_riak`` and ``is_sync``
        params.
        """
        persistence_helper = PersistenceHelper(use_riak=True, is_sync=True)
        self.assertEqual(persistence_helper.use_riak, True)
        self.assertEqual(persistence_helper.is_sync, True)

    def get_manager_inits(self):
        return (
            self._RiakManager.__init__,
            self._TxRiakManager.__init__,
            self._RedisManager.__init__,
            self._TxRedisManager.__init__,
        )

    def test_setup_applies_patches(self):
        """
        PersistenceHelper.setup() should apply patches to the persistence
        managers and return ``None``, not a Deferred.
        """
        manager_inits = self.get_manager_inits()
        persistence_helper = PersistenceHelper()
        self.assertEqual(persistence_helper._patches_applied, False)
        self.assertEqual(manager_inits, self.get_manager_inits())

        self.assertEqual(persistence_helper.setup(), None)
        self.assertEqual(persistence_helper._patches_applied, True)
        self.assertNotEqual(manager_inits, self.get_manager_inits())

        # Clean up after ourselves.
        persistence_helper._unpatch()
        self.assertEqual(persistence_helper._patches_applied, False)
        self.assertEqual(manager_inits, self.get_manager_inits())

    def test_cleanup_restores_patches(self):
        """
        PersistenceHelper.cleanup() should restore any patches applied by
        PersistenceHelper.setup().
        """
        manager_inits = self.get_manager_inits()
        persistence_helper = PersistenceHelper()
        self.assertEqual(persistence_helper.setup(), None)
        self.assertEqual(persistence_helper._patches_applied, True)
        self.assertNotEqual(manager_inits, self.get_manager_inits())

        success_result_of(persistence_helper.cleanup())
        self.assertEqual(persistence_helper._patches_applied, False)
        self.assertEqual(manager_inits, self.get_manager_inits())

    def test_get_riak_manager_unpatched(self):
        """
        .get_riak_manager() should fail if .setup() has not been called.
        """
        persistence_helper = PersistenceHelper()
        err = self.assertRaises(Exception, persistence_helper.get_riak_manager)
        self.assertTrue('setup() must be called' in str(err))

    def test_get_redis_manager_unpatched(self):
        """
        .get_redis_manager() should fail if .setup() has not been called.
        """
        persistence_helper = PersistenceHelper()
        err = self.assertRaises(
            Exception, persistence_helper.get_redis_manager)
        self.assertTrue('setup() must be called' in str(err))

    def test_mk_config_unpatched(self):
        """
        .mk_config() should fail if .setup() has not been called.
        """
        persistence_helper = PersistenceHelper()
        err = self.assertRaises(Exception, persistence_helper.mk_config, {})
        self.assertTrue('setup() must be called' in str(err))

    def test_get_riak_manager_no_riak(self):
        """
        .get_riak_manager() should fail if ``use_riak`` is ``False``.
        """
        persistence_helper = self.add_helper(PersistenceHelper())
        err = self.assertRaises(Exception, persistence_helper.get_riak_manager)
        self.assertTrue(
            'Use of Riak has been disabled for this test.' in str(err))

    def test_get_riak_manager_sync(self):
        """
        .get_riak_manager() should return a RiakManager if ``is_sync`` is
        ``True``.
        """
        persistence_helper = self.add_helper(
            PersistenceHelper(use_riak=True, is_sync=True))
        manager = persistence_helper.get_riak_manager()
        self.assertIsInstance(manager, self._RiakManager)
        self.assertEqual(persistence_helper._riak_managers, [manager])

    def test_get_riak_manager_async(self):
        """
        .get_riak_manager() should return a TxRiakManager if ``is_sync`` is
        ``False``.
        """
        persistence_helper = self.add_helper(PersistenceHelper(use_riak=True))
        manager = persistence_helper.get_riak_manager()
        self.assertIsInstance(manager, self._TxRiakManager)
        self.assertEqual(persistence_helper._riak_managers, [manager])

    def test_get_redis_manager_sync(self):
        """
        .get_redis_manager() should return a RedisManager if ``is_sync`` is
        ``True``.
        """
        persistence_helper = self.add_helper(PersistenceHelper(is_sync=True))
        manager = persistence_helper.get_redis_manager()
        self.assertIsInstance(manager, self._RedisManager)
        self.assertEqual(persistence_helper._redis_managers, [manager])

    @inlineCallbacks
    def test_get_redis_manager_async(self):
        """
        .get_redis_manager() should return a Deferred that fires with a
        TxRedisManager if ``is_sync`` is ``False``.
        """
        persistence_helper = self.add_helper(PersistenceHelper())
        manager_d = persistence_helper.get_redis_manager()
        self.assertIsInstance(manager_d, Deferred)
        manager = yield manager_d
        self.assertIsInstance(manager, self._TxRedisManager)
        self.assertEqual(persistence_helper._redis_managers, [manager])

    def test_mk_config(self):
        """
        .mk_config() should return a copy of the provided config with
        riak_manager and redis_manager fields overridden.
        """
        persistence_helper = self.add_helper(PersistenceHelper())
        config = {}
        new_config = persistence_helper.mk_config(config)
        self.assertEqual(
            ['redis_manager', 'riak_manager'], sorted(new_config.keys()))
        self.assertEqual(config, {})

    def test_deepcopy_RiakDisabledForTest(self):
        """
        RiakDisabledForTest should be deepcopy()able.
        """
        rdft = RiakDisabledForTest()
        self.assertEqual(rdft, deepcopy(rdft))

    def test__get_riak_managers_for_cleanup(self):
        """
        ._get_riak_managers_for_cleanup() should return the known Riak managers
        in reverse order with appropriate boolean determining whether they
        should be purged or not.
        """
        persistence_helper = PersistenceHelper()
        managers = [
            FakeRiakManagerForCleanup('bucket1'),
            FakeRiakManagerForCleanup('bucket2'),
            FakeRiakManagerForCleanup('bucket2'),
            FakeRiakManagerForCleanup('bucket1'),
            FakeRiakManagerForCleanup('bucket3'),
        ]
        persistence_helper._riak_managers.extend(managers)
        self.assertEqual(
            list(persistence_helper._get_riak_managers_for_cleanup()),
            list(reversed(zip([True, True, False, False, True], managers))))

    def test__get_redis_managers_for_cleanup(self):
        """
        ._get_redis_managers_for_cleanup() should return the known Redis
        managers in reverse order with appropriate boolean determining whether
        they should be purged or not.
        """
        persistence_helper = PersistenceHelper()
        managers = [
            FakeRedisManagerForCleanup('prefix1'),
            FakeRedisManagerForCleanup('prefix2'),
            FakeRedisManagerForCleanup('prefix2'),
            FakeRedisManagerForCleanup('prefix1'),
            FakeRedisManagerForCleanup('prefix3'),
        ]
        persistence_helper._redis_managers.extend(managers)
        self.assertEqual(
            list(persistence_helper._get_redis_managers_for_cleanup()),
            list(reversed(zip([True, True, False, False, True], managers))))

    def test__purge_riak(self):
        """
        ._purge_riak() should call manager.purge_all().
        """
        persistence_helper = PersistenceHelper()
        manager = FakeRiakManagerForCleanup('prefix1')
        self.assertEqual(manager.purged, False)
        self.assertEqual(
            persistence_helper._purge_riak(manager), 'maybe async')
        self.assertEqual(manager.purged, True)

    def test__purge_redis(self):
        """
        ._purge_redis() should call manager._purge_all() and remain connected.
        """
        persistence_helper = PersistenceHelper()
        manager = FakeRedisManagerForCleanup('prefix1')
        self.assertEqual(manager.purged, False)
        self.assertEqual(manager.connected, True)
        success_result_of(persistence_helper._purge_redis(manager))
        self.assertEqual(manager.purged, True)
        self.assertEqual(manager.connected, True)

    def test__purge_redis_not_connected(self):
        """
        ._purge_redis() should ignore riak managers that aren't connected.
        """
        persistence_helper = PersistenceHelper()
        manager = FakeRedisManagerForCleanup('prefix1')
        manager.close_manager()
        self.assertEqual(manager.purged, False)
        success_result_of(persistence_helper._purge_redis(manager))
        self.assertEqual(manager.purged, False)

    def test_cleanup_purges_managers(self):
        """
        .cleanup() should purge the Riak and Redis managers that need purging.
        """
        persistence_helper = PersistenceHelper()
        riak_purge = FakeRiakManagerForCleanup('bucket1')
        riak_nopurge = FakeRiakManagerForCleanup('bucket1')
        redis_purge = FakeRedisManagerForCleanup('prefix1')
        redis_nopurge = FakeRedisManagerForCleanup('prefix1')
        persistence_helper._riak_managers.extend([riak_purge, riak_nopurge])
        persistence_helper._redis_managers.extend([redis_purge, redis_nopurge])

        success_result_of(persistence_helper.cleanup())
        self.assertEqual(
            [True, False], [riak_purge.purged, riak_nopurge.purged])
        self.assertEqual(
            [True, False], [redis_purge.purged, redis_nopurge.purged])

    def test_cleanup_closes_riak_managers(self):
        """
        .cleanup() should close Riak client connections.
        """
        persistence_helper = PersistenceHelper()
        conn1 = FakeRiakClientConnection()
        conn2 = FakeRiakClientConnection()
        manager = FakeRiakManagerForCleanup('bucket1', [conn1, conn2])
        persistence_helper._riak_managers.append(manager)
        self.assertEqual(manager.fake_conns, [conn1, conn2])
        self.assertEqual([conn1.closed, conn2.closed], [False, False])
        success_result_of(persistence_helper.cleanup())
        self.assertEqual(manager.fake_conns, [])
        self.assertEqual([conn1.closed, conn2.closed], [True, True])



# -*- test-case-name: vumi.tests.test_utils -*-

import os.path
import re
import sys
import base64
import pkg_resources
import warnings
from functools import wraps

from zope.interface import implements
from twisted.internet import defer
from twisted.internet import reactor, protocol
from twisted.internet.defer import succeed
from twisted.python.failure import Failure
from twisted.web.client import Agent, ResponseDone, WebClientContextFactory
from twisted.web.server import Site
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from twisted.web.http import PotentialDataLoss
from twisted.web.resource import Resource

from vumi.errors import VumiError


# Stop Agent from logging two useless lines for every request.
# This is hacky, but there's no better way to do it for now.
from twisted.web import client
client._HTTP11ClientFactory.noisy = False


def import_module(name):
    """
    This is a simpler version of `importlib.import_module` and does
    not support relative imports.

    It's here so that we can avoid using importlib and not have to
    juggle different deps between Python versions.
    """
    __import__(name)
    return sys.modules[name]


def to_kwargs(kwargs):
    """
    Convert top-level keys from unicode to string for older Python versions.

    See http://bugs.python.org/issue2646 for details.
    """
    return dict((k.encode('utf8'), v) for k, v in kwargs.iteritems())


class HttpError(VumiError):
    """Base class for errors raised by http_request_full."""


class HttpDataLimitError(VumiError):
    """Returned by http_request_full if too much data is returned."""


class HttpTimeoutError(VumiError):
    """Returned by http_request_full if the request times out."""


class SimplishReceiver(protocol.Protocol):
    def __init__(self, response, data_limit=None):
        self.deferred = defer.Deferred(canceller=self.cancel_on_timeout)
        self.response = response
        self.data_limit = data_limit
        self.data_recvd_len = 0
        self.response.delivered_body = ''
        if response.code == 204:
            self.deferred.callback(self.response)
        else:
            response.deliverBody(self)

    def cancel_on_timeout(self, d):
        self.cancel_receiving(
            HttpTimeoutError("Timeout while receiving data")
        )

    def cancel_on_data_limit(self):
        self.cancel_receiving(
            HttpDataLimitError("More than %d bytes received"
                               % (self.data_limit,))
        )

    def cancel_receiving(self, err):
        self.transport.stopProducing()
        self.deferred.errback(err)

    def data_limit_exceeded(self):
        return (self.data_limit is not None and
                self.data_recvd_len > self.data_limit)

    def dataReceived(self, data):
        self.data_recvd_len += len(data)
        if self.data_limit_exceeded():
            self.cancel_on_data_limit()
        self.response.delivered_body += data

    def connectionLost(self, reason):
        if self.deferred.called:
            # this happens when the deferred is cancelled and this
            # triggers connection closing
            return
        if reason.check(ResponseDone):
            self.deferred.callback(self.response)
        elif reason.check(PotentialDataLoss):
            # This is only (and always!) raised if we have an HTTP 1.0 request
            # with no Content-Length.
            # See http://twistedmatrix.com/trac/ticket/4840 for sadness.
            #
            # We ignore this and treat the call as success. If we care about
            # checking for potential data loss, we should do that in all cases
            # rather than trying to figure out if we might need to.
            self.deferred.callback(self.response)
        else:
            self.deferred.errback(reason)


def http_request_full(url, data=None, headers={}, method='POST',
                      timeout=None, data_limit=None, context_factory=None,
                      agent_class=Agent):
    context_factory = context_factory or WebClientContextFactory()
    agent = agent_class(reactor, contextFactory=context_factory)
    d = agent.request(method,
                      url,
                      mkheaders(headers),
                      StringProducer(data) if data else None)

    def handle_response(response):
        return SimplishReceiver(response, data_limit).deferred

    d.addCallback(handle_response)

    if timeout is not None:
        cancelling_on_timeout = [False]

        def raise_timeout(reason):
            if not cancelling_on_timeout[0]:
                return reason
            return Failure(HttpTimeoutError("Timeout while connecting"))

        def cancel_on_timeout():
            cancelling_on_timeout[0] = True
            d.cancel()

        def cancel_timeout(r, delayed_call):
            if delayed_call.active():
                delayed_call.cancel()
            return r

        d.addErrback(raise_timeout)
        delayed_call = reactor.callLater(timeout, cancel_on_timeout)
        d.addCallback(cancel_timeout, delayed_call)

    return d


def mkheaders(headers):
    """
    Turn a dict of HTTP headers into an instance of Headers.

    Twisted expects a list of values, not a single value. We should
    support both.
    """
    raw_headers = {}
    for k, v in headers.iteritems():
        if isinstance(v, basestring):
            v = [v]
        raw_headers[k] = v
    return Headers(raw_headers)


def http_request(url, data, headers={}, method='POST'):
    d = http_request_full(url, data, headers=headers, method=method)
    return d.addCallback(lambda r: r.delivered_body)


def basic_auth_string(username, password):
    """
    Encode a username and password for use in an HTTP Basic Authentication
    header
    """
    b64 = base64.encodestring('%s:%s' % (username, password)).strip()
    return 'Basic %s' % b64


def normalize_msisdn(raw, country_code=''):
    # don't touch shortcodes
    if len(raw) <= 5:
        return raw

    raw = ''.join([c for c in raw if c.isdigit() or c == '+'])
    if raw.startswith('00'):
        return '+' + raw[2:]
    if raw.startswith('0'):
        return '+' + country_code + raw[1:]
    if raw.startswith('+'):
        return raw
    if raw.startswith(country_code):
        return '+' + raw
    return raw


class StringProducer(object):
    """
    For various twisted.web mechanics we need a producer to produce
    content for HTTP requests, this is a helper class to quickly
    create a producer for a bit of content
    """
    implements(IBodyProducer)

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


def build_web_site(resources, site_class=None):
    """Build a Twisted web Site instance for a specified dictionary of
    resources.

    :param dict resources:
        Dictionary of path -> resource class mappings to create the site from.
    :type site_class: Sub-class of Twisted's Site
    :param site_class:
        Site class to create. Defaults to :class:`LogFilterSite`.
    """
    if site_class is None:
        site_class = LogFilterSite

    root = Resource()
    # sort by ascending path length to make sure we create
    # resources lower down in the path earlier
    resources = resources.items()
    resources = sorted(resources, key=lambda r: len(r[0]))

    def create_node(node, path):
        if path in node.children:
            return node.children.get(path)
        else:
            new_node = Resource()
            node.putChild(path, new_node)
            return new_node

    for path, resource in resources:
        request_path = filter(None, path.split('/'))
        nodes, leaf = request_path[0:-1], request_path[-1]
        parent = reduce(create_node, nodes, root)
        parent.putChild(leaf, resource)

    site_factory = site_class(root)
    return site_factory


class LogFilterSite(Site):
    def log(self, request):
        if getattr(request, 'do_not_log', None):
            return
        return Site.log(self, request)


class PkgResources(object):
    """
    A helper for accessing a packages data files.

    :param str modname:
        The full dotted name of the module. E.g.
        ``vumi.resources``.
    """
    def __init__(self, modname):
        self.modname = modname

    def path(self, path):
        """
        Return the absolute path to a package resource.

        If path is already absolute, it is returned unmodified.

        :param str path:
            The relative or absolute path to the resource.
        """
        if os.path.isabs(path):
            return path
        return pkg_resources.resource_filename(self.modname, path)


vumi_resource_path = PkgResources("vumi.resources").path


def load_class(module_name, class_name):
    """
    Load a class when given its module and its class name

    >>> load_class('vumi.workers.example','ExampleWorker') # doctest: +ELLIPSIS
    <class vumi.workers.example.ExampleWorker at ...>
    >>>

    """
    mod = import_module(module_name)
    return getattr(mod, class_name)


def load_class_by_string(class_path):
    """
    Load a class when given its full name, including modules in python
    dot notation

    >>> cls = 'vumi.workers.example.ExampleWorker'
    >>> load_class_by_string(cls) # doctest: +ELLIPSIS
    <class vumi.workers.example.ExampleWorker at ...>
    >>>

    """
    parts = class_path.split('.')
    module_name = '.'.join(parts[:-1])
    class_name = parts[-1]
    return load_class(module_name, class_name)


def redis_from_config(redis_config):
    """
    Return a redis client instance from a config.

    If redis_config:

    * equals 'FAKE_REDIS', a new instance of :class:`FakeRedis` is returned.
    * is an instance of :class:`FakeRedis` that instance is returned

    Otherwise a new real redis client is returned.
    """
    warnings.warn("Use of redis directly is deprecated. Use vumi.persist "
                  "instead.", category=DeprecationWarning)

    import redis
    from vumi.persist import fake_redis
    if redis_config == "FAKE_REDIS":
        return fake_redis.FakeRedis()
    if isinstance(redis_config, fake_redis.FakeRedis):
        return redis_config
    return redis.Redis(**redis_config)


def flatten_generator(generator_func):
    """
    This is a synchronous version of @inlineCallbacks.

    NOTE: It doesn't correctly handle returnValue() being called in a
    non-decorated function called from the function we're decorating. We could
    copy the Twisted code to do that, but it's messy.
    """
    @wraps(generator_func)
    def wrapped(*args, **kw):
        gen = generator_func(*args, **kw)
        result = None
        while True:
            try:
                result = gen.send(result)
            except StopIteration:
                # Fell off the end, or "return" statement.
                return None
            except defer._DefGen_Return, e:
                # returnValue() called.
                return e.value

    return wrapped


def filter_options_on_prefix(options, prefix, delimiter='-'):
    """
    splits an options dict based on key prefixes

    >>> filter_options_on_prefix({'foo-bar-1': 'ok'}, 'foo')
    {'bar-1': 'ok'}
    >>>

    """
    return dict((key.split(delimiter, 1)[1], value)
                for key, value in options.items()
                if key.startswith(prefix))


def get_first_word(content, delimiter=' '):
    """
    Returns the first word from a string.

    Example::

      >>> get_first_word('KEYWORD rest of message')
      'KEYWORD'

    :type content: str or None
    :param content:
        Content from which the first word will be retrieved. If the
        content is None it is treated as an empty string (this is a
        convenience for dealing with content-less messages).
    :param str delimiter:
        Delimiter to split the string on. Default is ' '.
        Passed to :func:`string.partition`.
    :returns:
        A string containing the first word.
    """
    return (content or '').partition(delimiter)[0]


def cleanup_msisdn(number, country_code):
    number = re.sub('\+', '', number)
    number = re.sub('^0', country_code, number)
    return number


def get_operator_name(msisdn, mapping):
    for key, value in mapping.items():
        if msisdn.startswith(str(key)):
            if isinstance(value, dict):
                return get_operator_name(msisdn, value)
            return value
    return 'UNKNOWN'


def get_operator_number(msisdn, country_code, mapping, numbers):
    msisdn = cleanup_msisdn(msisdn, country_code)
    operator = get_operator_name(msisdn, mapping)
    number = numbers.get(operator)
    return number


def safe_routing_key(routing_key):
    """
    >>> safe_routing_key(u'*32323#')
    u's32323h'
    >>>

    """
    return reduce(lambda r_key, kv: r_key.replace(*kv),
                    [('*', 's'), ('#', 'h')], routing_key)


def generate_worker_id(system_id, worker_id):
    return "%s:%s" % (system_id, worker_id,)



from vumi.service import Worker
from vumi.tests.utils import get_stubbed_worker, LogCatcher
from vumi.tests.fake_amqp import FakeAMQClient
from vumi import log
from vumi.tests.helpers import VumiTestCase


class ToyWorker(Worker):
    def poke(self):
        return "poke"


class TestTestUtils(VumiTestCase):

    def test_get_stubbed_worker(self):
        worker = get_stubbed_worker(ToyWorker)
        self.assertEqual("poke", worker.poke())
        self.assertTrue(isinstance(worker._amqp_client, FakeAMQClient))

    def test_get_stubbed_worker_with_config(self):
        options = {'key': 'value'}
        worker = get_stubbed_worker(ToyWorker, options)
        self.assertEqual({}, worker._amqp_client.vumi_options)
        self.assertEqual(options, worker.config)


class TestLogCatcher(VumiTestCase):
    def test_simple_catching(self):
        lc = LogCatcher()
        with lc:
            log.info("Test")
        self.assertEqual(lc.messages(), ["Test"])

    def test_system_filtering(self):
        lc = LogCatcher(system="^ab")
        with lc:
            log.info("Test 1", system="abc")
            log.info("Test 2", system="def")
        self.assertEqual(lc.messages(), ["Test 1"])

    def test_message_filtering(self):
        lc = LogCatcher(message="^Keep")
        with lc:
            log.info("Keep this")
            log.info("Discard this")
        self.assertEqual(lc.messages(), ["Keep this"])

    def test_message_concatenation(self):
        lc = LogCatcher()
        with lc:
            log.info("Part 1", "Part 2")
        self.assertEqual(lc.messages(), ["Part 1 Part 2"])



# -*- test-case-name: vumi.transports.api.tests.test_oldapi -*-
import json
import re
from base64 import b64decode

from twisted.python import log
from twisted.web import http

from vumi.transports.httprpc import HttpRpcTransport


class OldSimpleHttpTransport(HttpRpcTransport):
    """
    Maintains the API used by the old Django based
    method of loading SMS's into VUMI over HTTP

    Configuration options:

    web_path : str
        The path relative to the host where this listens
    web_port : int
        The port this listens on
    transport_name : str
        The name this transport instance will use to create it's queues
    identities : dictionary
        user : str
        password : str
        default_transport : str
    """

    def validate_config(self):
        super(OldSimpleHttpTransport, self).validate_config()
        self.identities = self.config.get('identities', {})

    def get_health_response(self):
        return json.dumps({})

    def get_credentials(self, request):
        auth_header = 'Authorization'
        headers = request.requestHeaders
        if headers.hasHeader(auth_header):
            auth = headers.getRawHeaders(auth_header)[0]
            creds = b64decode(auth.split(' ')[-1])
            return creds.split(':')
        else:
            return '', ''

    def is_authorized(self, username, password):
        return self.identities.get(username) == password

    def handle_outbound_message(self, message):
        log.msg("OldSimpleHttpTransport consuming %s" % (message))
        return self.publish_ack(user_message_id=message['message_id'],
            sent_message_id=message['message_id'])

    def check_authorization(self, request):
        username, password = self.get_credentials(request)
        if self.identities and not self.is_authorized(username, password):
            return False, username
        return True, username

    def handle_raw_inbound_message(self, request_id, request):
        authorized, username = self.check_authorization(request)
        if not authorized:
            return self.finish_request(request_id, 'Not Authorized',
                                        code=http.UNAUTHORIZED)
        message = request.args.get('message', [None])[0]
        to_msisdns = request.args.get('to_msisdn', [])
        from_msisdn = request.args.get('from_msisdn', [None])[0]
        return_list = []
        for to_msisdn in to_msisdns:
            message_id = self.generate_message_id()
            content = message
            to_addr = to_msisdn
            from_addr = from_msisdn
            log.msg(
                'OldSimpleHttpTransport sending from %s to %s message "%s"' % (
                    from_addr, to_addr, content))

            self.publish_message(
                message_id=message_id,
                content=content,
                to_addr=to_addr,
                from_addr=from_addr,
                provider='vumi',
                transport_type='old_simple_http',
                transport_metadata={
                    'http_user': username,
                }
            )
            return_list.append({
                "message": message,
                "to_msisdn": to_msisdn,
                "from_msisdn": from_msisdn,
                "id": message_id,
                })
        return self.finish_request(request_id, json.dumps(return_list))


class OldTemplateHttpTransport(OldSimpleHttpTransport):

    def handle_outbound_message(self, message):
        log.msg("OldTemplateHttpTransport consuming %s" % (message))

    def extract_template_args(self, args, length):
        template_args = []
        for i in range(length):
            template_args.append({})
        for k, v in args.items():
            if k.startswith("template_"):
                for i, x in enumerate(v):
                    template_args[i][k] = x
        return template_args

    def handle_raw_inbound_message(self, request_id, request):
        authorized, username = self.check_authorization(request)
        if not authorized:
            return self.finish_request(request_id, 'Not Authorized',
                                        code=http.UNAUTHORIZED)
        opener = re.compile('{{ *')
        closer = re.compile(' *}}')
        template = request.args.get('template', [None])[0]
        template = opener.sub('%(template_', template)
        template = closer.sub(')s', template)
        to_msisdns = request.args.get('to_msisdn', [])
        from_msisdn = request.args.get('from_msisdn', [None])[0]
        template_args = self.extract_template_args(request.args,
                                                   len(to_msisdns))
        return_list = []
        for i, to_msisdn in enumerate(to_msisdns):
            message_id = self.generate_message_id()
            message = content = template % template_args[i]
            to_addr = to_msisdn
            from_addr = from_msisdn
            log.msg(('OldTemplateHttpTransport sending from %s to %s '
                     'message "%s"') % (from_addr, to_addr, content))
            self.publish_message(
                message_id=message_id,
                content=content,
                to_addr=to_addr,
                from_addr=from_addr,
                provider='vumi',
                transport_type='old_template_http',
                transport_metadata={
                    'http_user': username,
                }
            )
            return_list.append({
                "message": message,
                "to_msisdn": to_msisdn,
                "from_msisdn": from_msisdn,
                "id": message_id,
                })
        return self.finish_request(request_id, json.dumps(return_list))



# -*- coding: utf-8 -*-
from vumi.codecs.ivumi_codecs import IVumiCodec
from vumi.codecs.vumi_codecs import VumiCodec, VumiCodecException

from twisted.trial.unittest import TestCase


class TestVumiCodec(TestCase):

    def setUp(self):
        self.codec = VumiCodec()

    def test_implements(self):
        self.assertTrue(IVumiCodec.implementedBy(VumiCodec))
        self.assertTrue(IVumiCodec.providedBy(self.codec))

    def test_unicode_encode_guard(self):
        self.assertRaises(
            VumiCodecException, self.codec.encode, "byte string")

    def test_bytestring_decode_guard(self):
        self.assertRaises(
            VumiCodecException, self.codec.decode, u"unicode")

    def test_default_encoding(self):
        self.assertEqual(self.codec.encode(u"a"), "a")
        self.assertRaises(
            UnicodeEncodeError, self.codec.encode, u"ë")

    def test_default_decoding(self):
        self.assertEqual(self.codec.decode("a"), u"a")
        self.assertRaises(
            UnicodeDecodeError, self.codec.decode, '\xc3\xab')  # e-umlaut

    def test_encode_utf8(self):
        self.assertEqual(self.codec.encode(u"Zoë", "utf-8"), 'Zo\xc3\xab')

    def test_decode_utf8(self):
        self.assertEqual(self.codec.decode('Zo\xc3\xab', "utf-8"), u"Zoë")

    def test_encode_utf16be(self):
        self.assertEqual(
            self.codec.encode(u"Zoë", "utf-16be"), '\x00Z\x00o\x00\xeb')

    def test_decode_utf16be(self):
        self.assertEqual(
            self.codec.decode('\x00Z\x00o\x00\xeb', "utf-16be"), u"Zoë")

    def test_encode_ucs2(self):
        self.assertEqual(
            self.codec.encode(u"Zoë", "ucs2"), '\x00Z\x00o\x00\xeb')

    def test_decode_ucs2(self):
        self.assertEqual(
            self.codec.decode('\x00Z\x00o\x00\xeb', "ucs2"), u"Zoë")

    def test_encode_gsm0338(self):
        self.assertEqual(
            self.codec.encode(u"HÜLK", "gsm0338"),
            ''.join([chr(code) for code in [72, 94, 76, 75]]))

    def test_encode_gsm0338_extended(self):
        self.assertEqual(
            self.codec.encode(u"foo €", "gsm0338"),
            ''.join([chr(code) for code in [102, 111, 111, 32, 27, 101]]))

    def test_decode_gsm0338_extended(self):
        self.assertEqual(
            self.codec.decode(
                ''.join([chr(code) for code in [102, 111, 111, 32, 27, 101]]),
                'gsm0338'),
            u"foo €")

    def test_encode_gsm0338_strict(self):
        self.assertRaises(
            UnicodeEncodeError, self.codec.encode, u'Zoë', 'gsm0338')

    def test_encode_gsm0338_ignore(self):
        self.assertEqual(
            self.codec.encode(u"Zoë", "gsm0338", 'ignore'), 'Zo')

    def test_encode_gsm0338_replace(self):
        self.assertEqual(
            self.codec.encode(u"Zoë", "gsm0338", 'replace'), 'Zo?')

    def test_decode_gsm0338_strict(self):
        self.assertRaises(
            UnicodeDecodeError, self.codec.decode,
            u'Zoë'.encode('utf-8'), 'gsm0338')

    def test_decode_gsm0338_ignore(self):
        self.assertEqual(
            self.codec.decode(
                u'Zoë'.encode('utf-8'), "gsm0338", 'ignore'), u'Zo')

    def test_decode_gsm0338_replace(self):
        self.assertEqual(
            self.codec.decode(
                u'Zoë'.encode('utf-8'), "gsm0338", 'replace'), u'Zo??')



# -*- test-case-name: vumi.transports.mxit.tests.test_mxit -*-
import json
import base64
from urllib import urlencode, unquote_plus
from HTMLParser import HTMLParser

from twisted.web import http
from twisted.internet.defer import inlineCallbacks, returnValue

from vumi.config import ConfigText, ConfigInt, ConfigDict, ConfigList
from vumi.persist.txredis_manager import TxRedisManager
from vumi.transports.httprpc import HttpRpcTransport
from vumi.transports.mxit.responses import MxitResponse
from vumi.utils import http_request_full


class MxitTransportException(Exception):
    """Raised when the Mxit API returns an error"""


class MxitTransportConfig(HttpRpcTransport.CONFIG_CLASS):

    client_id = ConfigText(
        'The OAuth2 ClientID assigned to this transport.', required=True,
        static=True)
    client_secret = ConfigText(
        'The OAuth2 ClientSecret assigned to this transport.', required=True,
        static=True)
    timeout = ConfigInt(
        'Timeout for outbound Mxit HTTP API calls.', required=False,
        default=30, static=True)
    redis_manager = ConfigDict(
        'How to connect to Redis', required=True, static=True)
    api_send_url = ConfigText(
        'The URL for the Mxit message sending API.',
        required=False, default="https://api.mxit.com/message/send/",
        static=True)
    api_auth_url = ConfigText(
        'The URL for the Mxit authentication API.',
        required=False, default='https://auth.mxit.com',
        static=True)
    api_auth_scopes = ConfigList(
        'The list of scopes to request access to.',
        required=False, static=True, default=['message/send'])


class MxitTransport(HttpRpcTransport):
    """
    HTTP Transport for MXit, implemented using the MXit Mobi Portal
    (for inbound messages and replies) and the Messaging API (for sends
    that aren't replies).

    * Mobi Portal API specification:
      http://dev.mxit.com/docs/mobi-portal-api
    * Message API specification:
      https://dev.mxit.com/docs/restapi/messaging/post-message-send
    """

    CONFIG_CLASS = MxitTransportConfig
    content_type = 'text/html; charset=utf-8'
    transport_type = 'mxit'
    access_token_key = 'access_token'
    access_token_auto_decay = 0.95

    @inlineCallbacks
    def setup_transport(self):
        yield super(MxitTransport, self).setup_transport()
        config = self.get_static_config()
        self.redis = yield TxRedisManager.from_config(config.redis_manager)

    def is_mxit_request(self, request):
        return request.requestHeaders.hasHeader('X-Mxit-Contact')

    def noop(self, key):
        return key

    def parse_location(self, location):
        return dict(zip([
            'country_code',
            'country_name',
            'subdivision_code',
            'subdivision_name',
            'city_code',
            'city',
            'network_operator_id',
            'client_features_bitset',
            'cell_id'
        ], location.split(',')))

    def parse_profile(self, profile):
        return dict(zip([
            'language_code',
            'country_code',
            'date_of_birth',
            'gender',
            'tariff_plan',
        ], profile.split(',')))

    def html_decode(self, html):
        """
        Turns '&lt;b&gt;foo&lt;/b&gt;' into u'<b>foo</b>'
        """
        return HTMLParser().unescape(html)

    def get_request_data(self, request):
        headers = request.requestHeaders
        header_ops = [
            ('X-Device-User-Agent', self.noop),
            ('X-Mxit-Contact', self.noop),
            ('X-Mxit-USERID-R', self.noop),
            ('X-Mxit-Nick', self.noop),
            ('X-Mxit-Location', self.parse_location),
            ('X-Mxit-Profile', self.parse_profile),
            ('X-Mxit-User-Input', self.html_decode),
        ]
        data = {}
        for header, proc in header_ops:
            if headers.hasHeader(header):
                [value] = headers.getRawHeaders(header)
                data[header] = proc(value)
        return data

    def get_request_content(self, request):
        headers = request.requestHeaders
        [content] = headers.getRawHeaders('X-Mxit-User-Input', [None])
        if content:
            return unquote_plus(content)

        if request.args and 'input' in request.args:
            [content] = request.args['input']
            return content

        return None

    def handle_raw_inbound_message(self, msg_id, request):
        if not self.is_mxit_request(request):
            return self.finish_request(
                msg_id, data=http.RESPONSES[http.BAD_REQUEST],
                code=http.BAD_REQUEST)

        data = self.get_request_data(request)
        content = self.get_request_content(request)
        return self.publish_message(
            message_id=msg_id,
            content=content,
            to_addr=data['X-Mxit-Contact'],
            from_addr=data['X-Mxit-USERID-R'],
            provider='mxit',
            transport_type=self.transport_type,
            helper_metadata={
                'mxit_info': data,
            })

    def handle_outbound_message(self, message):
        self.emit("MxitTransport consuming %s" % (message))
        if message["in_reply_to"] is None:
            return self.handle_outbound_send(message)
        else:
            return self.handle_outbound_reply(message)

    @inlineCallbacks
    def handle_outbound_reply(self, message):
        missing_fields = self.ensure_message_values(
            message, ['in_reply_to'])
        if missing_fields:
            yield self.reject_message(message, missing_fields)
        else:
            yield self.render_response(message)
            yield self.publish_ack(
                user_message_id=message['message_id'],
                sent_message_id=message['message_id'])

    @inlineCallbacks
    def get_access_token(self):
        access_token = yield self.redis.get(self.access_token_key)
        if access_token is None:
            access_token, expiry = yield self.request_new_access_token()
            # always make sure we expire before the token actually does
            safe_expiry = expiry * self.access_token_auto_decay
            yield self.redis.setex(
                self.access_token_key, int(safe_expiry), access_token)
        returnValue(access_token)

    @inlineCallbacks
    def request_new_access_token(self):
        config = self.get_static_config()
        url = '%s/token' % (config.api_auth_url)
        auth = base64.b64encode(
            '%s:%s' % (config.client_id, config.client_secret))
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Basic %s' % (auth,)
        }
        data = urlencode({
            'grant_type': 'client_credentials',
            'scope': ' '.join(config.api_auth_scopes)
        })
        response = yield http_request_full(url=url, method='POST',
                                           headers=headers, data=data)
        data = json.loads(response.delivered_body)
        if 'error' in data:
            raise MxitTransportException(
                '%(error)s: %(error_description)s.' % data)

        returnValue(
            (data['access_token'].encode('utf8'), int(data['expires_in'])))

    @inlineCallbacks
    def handle_outbound_send(self, message):
        config = self.get_static_config()
        body = message['content']
        access_token = yield self.get_access_token()
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer %s" % (access_token,)
        }
        data = {
            "Body": body,
            "ContainsMarkup": "true",
            "From": message["from_addr"],
            "To": message["to_addr"],
            "Spool": "true",
        }
        context_factory = None
        resp = yield http_request_full(
            config.api_send_url, data=json.dumps(data), headers=headers,
            method="POST", timeout=config.timeout,
            context_factory=context_factory)

    @inlineCallbacks
    def render_response(self, message):
        msg_id = message['in_reply_to']
        request = self.get_request(msg_id)
        if request:
            data = yield MxitResponse(message).flatten()
            super(MxitTransport, self).finish_request(
                msg_id, data, code=http.OK)



# -*- test-case-name: vumi.transports.airtel.tests.test_airtel -*-

import json
import re

from twisted.internet.defer import inlineCallbacks
from twisted.web import http

from vumi.transports.httprpc import HttpRpcTransport
from vumi.components.session import SessionManager
from vumi.message import TransportUserMessage
from vumi import log
from vumi.config import ConfigInt, ConfigText, ConfigBool, ConfigDict


class AirtelUSSDTransportConfig(HttpRpcTransport.CONFIG_CLASS):
    airtel_username = ConfigText('The username for this transport',
                                 default=None, static=True)
    airtel_password = ConfigText('The password for this transport',
                                 default=None, static=True)
    airtel_charge = ConfigBool(
        'Whether or not to charge for the responses sent.', required=False,
        default=False, static=True)
    airtel_charge_amount = ConfigInt('How much to charge', default=0,
                                     required=False, static=True)
    redis_manager = ConfigDict('Parameters to connect to Redis with.',
                               default={}, required=False, static=True)
    session_key_prefix = ConfigText(
        'The prefix to use for session key management. Specify this'
        'if you are using more than 1 worker in a load-balanced'
        'fashion.', default=None, static=True)
    ussd_session_timeout = ConfigInt('Max length of a USSD session',
                                     default=60 * 10, required=False,
                                     static=True)
    to_addr_pattern = ConfigText(
        'A regular expression that to_addr values in messages that start a'
        ' new USSD session must match. Initial messages with invalid'
        ' to_addr values are rejected.',
        default=None, required=False, static=True,
    )


class AirtelUSSDTransport(HttpRpcTransport):
    """
    Client implementation for the Comviva Flares HTTP Pull API.
    Based on Flares 1.5.0, document version 1.2.0
    """

    transport_type = 'ussd'
    content_type = 'text/plain; charset=utf-8'
    to_addr_re = None
    ENCODING = 'utf-8'
    CONFIG_CLASS = AirtelUSSDTransportConfig
    EXPECTED_AUTH_FIELDS = set(['userid', 'password'])
    EXPECTED_CLEANUP_FIELDS = set(['SessionID', 'msisdn', 'clean', 'error'])
    EXPECTED_USSD_FIELDS = set(['SessionID', 'MSISDN', 'MSC', 'input'])

    @inlineCallbacks
    def setup_transport(self):
        super(AirtelUSSDTransport, self).setup_transport()
        config = self.get_static_config()
        self.session_manager = yield SessionManager.from_redis_config(
            config.redis_manager, self.get_session_key_prefix(),
            config.ussd_session_timeout)
        if config.to_addr_pattern is not None:
            self.to_addr_re = re.compile(config.to_addr_pattern)

    def get_session_key_prefix(self):
        config = self.get_static_config()
        default_session_key_prefix = "vumi.transports.airtel:%s" % (
                                        self.transport_name,)
        return (config.session_key_prefix or default_session_key_prefix)

    def is_cleanup(self, request):
        return 'clean' in request.args

    def requires_auth(self):
        config = self.get_static_config()
        return (None not in (config.airtel_username, config.airtel_password))

    def is_authenticated(self, request):
        config = self.get_static_config()
        if self.EXPECTED_AUTH_FIELDS.issubset(request.args):
            username = request.args['userid'][0]
            password = request.args['password'][0]
            auth = (username == config.airtel_username and
                    password == config.airtel_password)
            if not auth:
                log.msg('Invalid authentication credentials: %s:%s' % (
                        username, password))
            return auth

    def valid_to_addr(self, to_addr):
        if self.to_addr_re is None:
            return True
        return bool(self.to_addr_re.match(to_addr))

    def handle_bad_request(self, message_id, request, errors):
        log.msg('Unhappy incoming message: %s' % (errors,))
        return self.finish_request(message_id, json.dumps(errors),
                                   code=http.BAD_REQUEST)

    def handle_raw_inbound_message(self, message_id, request):
        if self.requires_auth() and not self.is_authenticated(request):
            self.finish_request(message_id, 'Forbidden', code=http.FORBIDDEN)
            return

        if self.is_cleanup(request):
            return self.handle_cleanup_request(message_id, request)
        return self.handle_ussd_request(message_id, request)

    @inlineCallbacks
    def handle_cleanup_request(self, message_id, request):
        if self.requires_auth():
            fields = self.EXPECTED_CLEANUP_FIELDS.union(
                self.EXPECTED_AUTH_FIELDS)
        else:
            fields = self.EXPECTED_CLEANUP_FIELDS

        values, errors = self.get_field_values(request, fields)
        if errors:
            self.handle_bad_request(message_id, request, errors)
            return

        session_id = values['SessionID']
        session = yield self.session_manager.load_session(session_id)
        if not session:
            log.warning('Received cleanup for unknown airtel session.',
                        session_id=session_id)
            self.finish_request(message_id, 'Unknown Session', code=http.OK)
            return

        from_addr = values['msisdn']
        to_addr = session['to_addr']
        session_event = TransportUserMessage.SESSION_CLOSE
        yield self.session_manager.clear_session(session_id)
        yield self.publish_message(
            message_id=message_id,
            content='',
            to_addr=to_addr,
            from_addr=from_addr,
            provider='airtel',
            session_event=session_event,
            transport_type=self.transport_type,
            transport_metadata={
                'airtel': {
                    'clean': values['clean'],
                    'error': values['error'],
                },
            })
        self.finish_request(message_id, '', code=http.OK)

    @inlineCallbacks
    def handle_ussd_request(self, message_id, request):
        if self.requires_auth():
            fields = self.EXPECTED_USSD_FIELDS.union(
                self.EXPECTED_AUTH_FIELDS)
        else:
            fields = self.EXPECTED_USSD_FIELDS

        values, errors = self.get_field_values(request, fields)
        if errors:
            self.handle_bad_request(message_id, request, errors)
            return

        session_id = values['SessionID']
        from_addr = values['MSISDN']

        session = yield self.session_manager.load_session(session_id)
        if session:
            to_addr = session['to_addr']
            yield self.session_manager.save_session(session_id, session)
            session_event = TransportUserMessage.SESSION_RESUME
            content = values['input']
        else:
            # Airtel doesn't provide us with the full to_addr, the start *
            # and ending # are omitted, add those again so we can use it
            # for internal routing.
            to_addr = '*%s#' % (values['input'],)
            if self.valid_to_addr(to_addr):
                yield self.session_manager.create_session(
                    session_id, from_addr=from_addr, to_addr=to_addr)
                session_event = TransportUserMessage.SESSION_NEW
                content = ''
            else:
                self.handle_bad_request(message_id, request, {
                    "invalid_session": (
                        "Session id %r has not been encountered in the last %s"
                        " seconds and the 'input' request parameter value"
                        " %r doesn't look like a valid USSD address."
                        % (session_id, self.session_manager.max_session_length,
                           to_addr)
                    )
                })
                return

        yield self.publish_message(
            message_id=message_id,
            content=content,
            to_addr=to_addr,
            from_addr=from_addr,
            provider='airtel',
            session_event=session_event,
            transport_type=self.transport_type,
            transport_metadata={
                'airtel': {
                    'MSC': values['MSC'],
                },
            })

    def handle_outbound_message(self, message):
        config = self.get_static_config()
        missing_fields = self.ensure_message_values(
            message, ['in_reply_to', 'content'])

        if missing_fields:
            return self.reject_message(message, missing_fields)

        if message['session_event'] == TransportUserMessage.SESSION_CLOSE:
            free_flow = 'FB'
        else:
            free_flow = 'FC'

        headers = {
            'Freeflow': [free_flow],
            'charge': [('Y' if config.airtel_charge else 'N')],
            'amount': [str(config.airtel_charge_amount)],
        }

        content = message['content'].encode(self.ENCODING).lstrip()

        if self.noisy:
            log.debug('in_reply_to: %s' % (message['in_reply_to'],))
            log.debug('content: %r' % (content,))
            log.debug('Response headers: %r' % (headers,))

        self.finish_request(
            message['in_reply_to'],
            content,
            code=http.OK,
            headers=headers)
        return self.publish_ack(
            user_message_id=message['message_id'],
            sent_message_id=message['message_id'])



from collections import namedtuple
from functools import partial

from twisted.internet.defer import succeed
from twisted.trial.unittest import TestCase
from twisted.web import http

from vumi.transports.parlayx.soaputil import (
    SOAP_ENV, soap_envelope, unwrap_soap_envelope, soap_fault, tostring,
    perform_soap_request, SoapFault)
from vumi.transports.parlayx.xmlutil import (
    ParseError, gettext, Element, LocalNamespace as L, element_to_dict)
from vumi.transports.parlayx.tests.utils import (
    MockResponse, _FailureResultOfMixin)


class SoapWrapperTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.soaputil.soap_envelope`,
    `vumi.transports.parlayx.soaputil.unwrap_soap_envelope` and
    `vumi.transports.parlayx.soaputil.soap_fault`.
    """
    def test_soap_envelope(self):
        """
        `soap_envelope` wraps content in a SOAP envelope element.
        """
        self.assertEqual(
            '<soapenv:Envelope xmlns:soapenv='
            '"http://schemas.xmlsoap.org/soap/envelope/"><soapenv:Body>'
            'hello</soapenv:Body></soapenv:Envelope>',
            tostring(soap_envelope('hello')))
        self.assertEqual(
            '<soapenv:Envelope xmlns:soapenv='
            '"http://schemas.xmlsoap.org/soap/envelope/"><soapenv:Body>'
            '<tag>hello</tag></soapenv:Body></soapenv:Envelope>',
            tostring(soap_envelope(Element('tag', 'hello'))))

    def test_unwrap_soap_envelope(self):
        """
        `unwrap_soap_envelope` unwraps a SOAP envelope element, with no header,
        to a tuple of the SOAP body element and ``None``.
        """
        body, header = unwrap_soap_envelope(
            soap_envelope(Element('tag', 'hello')))
        self.assertIdentical(None, header)
        self.assertEqual(
            '<soapenv:Body xmlns:soapenv='
            '"http://schemas.xmlsoap.org/soap/envelope/"><tag>hello</tag>'
            '</soapenv:Body>',
            tostring(body))

    def test_unwrap_soap_envelope_header(self):
        """
        `unwrap_soap_envelope` unwraps a SOAP envelope element, with a header,
        to a tuple of the SOAP body and header elements.
        """
        body, header = unwrap_soap_envelope(
            soap_envelope(
                Element('tag', 'hello'),
                Element('header', 'value')))
        self.assertEqual(
            '<soapenv:Header xmlns:soapenv='
            '"http://schemas.xmlsoap.org/soap/envelope/">'
            '<header>value</header></soapenv:Header>',
            tostring(header))
        self.assertEqual(
            '<soapenv:Body xmlns:soapenv='
            '"http://schemas.xmlsoap.org/soap/envelope/">'
            '<tag>hello</tag></soapenv:Body>',
            tostring(body))

    def test_soap_fault(self):
        """
        `soap_fault` constructs a SOAP fault element from a code and
        description.
        """
        self.assertEqual(
            '<soapenv:Fault xmlns:soapenv='
            '"http://schemas.xmlsoap.org/soap/envelope/">'
            '<faultcode>soapenv:Client</faultcode>'
            '<faultstring>Oops.</faultstring></soapenv:Fault>',
            tostring(soap_fault('soapenv:Client', 'Oops.')))


class ToyFaultDetail(namedtuple('ToyFaultDetail', ['foo', 'bar'])):
    """
    A SOAP fault detail used for tests.
    """
    @classmethod
    def from_element(cls, elem):
        if elem.tag != L.ToyFaultDetail.text:
            return None
        return cls(gettext(elem, 'foo'), gettext(elem, 'bar'))


class TestFault(SoapFault):
    """
    A SOAP fault used for tests.
    """
    detail_type = ToyFaultDetail


def _make_fault(*a, **kw):
    """
    Create a SOAP body containing a SOAP fault.
    """
    return SOAP_ENV.Body(soap_fault(*a, **kw))


class SoapFaultTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.soaputil.SoapFault`.
    """
    def test_missing_fault(self):
        """
        `SoapFault.from_element` raises `ValueError` if the element contains no
        SOAP fault.
        """
        self.assertRaises(ValueError,
            SoapFault.from_element, Element('tag'))

    def test_from_element(self):
        """
        `SoapFault.from_element` creates a `SoapFault` instance from an
        ElementTree element and parses known SOAP fault details.
        """
        detail = L.ToyFaultDetail(
                L.foo('a'),
                L.bar('b'))

        fault = SoapFault.from_element(_make_fault(
            'soapenv:Client', 'message', 'actor', detail=detail))
        self.assertEqual(
            ('soapenv:Client', 'message', 'actor'),
            (fault.code, fault.string, fault.actor))
        self.assertIdentical(None, fault.parsed_detail)

    def test_to_element(self):
        """
        `SoapFault.to_element` serializes the fault to a SOAP ``Fault``
        ElementTree element.
        """
        detail = L.ToyFaultDetail(
                L.foo('a'),
                L.bar('b'))

        fault = SoapFault.from_element(_make_fault(
            'soapenv:Client', 'message', 'actor', detail=detail))
        self.assertEqual(
            {str(SOAP_ENV.Fault): {
                'faultcode': fault.code,
                'faultstring': fault.string,
                'faultactor': fault.actor,
                'detail': {
                    'ToyFaultDetail': {'foo': 'a',
                                        'bar': 'b'}}}},
            element_to_dict(fault.to_element()))

    def test_to_element_no_detail(self):
        """
        `SoapFault.to_element` serializes the fault to a SOAP ``Fault``
        ElementTree element, omitting the ``detail`` element if
        `SoapFault.detail` is None.
        """
        fault = SoapFault.from_element(_make_fault(
            'soapenv:Client', 'message', 'actor'))
        self.assertEqual(
            {str(SOAP_ENV.Fault): {
                'faultcode': fault.code,
                'faultstring': fault.string,
                'faultactor': fault.actor}},
            element_to_dict(fault.to_element()))

    def test_expected_faults(self):
        """
        `SoapFault.from_element` creates an instance of a specified `SoapFault`
        subclass if a fault detail of a recognised type occurs.
        """
        detail = [
            L.WhatIsThis(
                L.foo('a'),
                L.bar('b')),
            L.ToyFaultDetail(
                L.foo('c'),
                L.bar('d'))]

        fault = SoapFault.from_element(_make_fault(
            'soapenv:Client', 'message', 'actor', detail=detail),
            expected_faults=[TestFault])
        self.assertEqual(
            ('soapenv:Client', 'message', 'actor'),
            (fault.code, fault.string, fault.actor))
        parsed_detail = fault.parsed_detail
        self.assertEqual(
            ('c', 'd'),
            (parsed_detail.foo, parsed_detail.bar))


class PerformSoapRequestTests(_FailureResultOfMixin, TestCase):
    """
    Tests for `vumi.transports.parlayx.soaputil.perform_soap_request`.
    """
    def setUp(self):
        self.requests = []

    def _http_request_full(self, response, uri, body, headers):
        """
        A mock for `vumi.utils.http_request_full`.

        Store an HTTP request's information and return a canned response.
        """
        self.requests.append((uri, body, headers))
        return succeed(response)

    def _perform_soap_request(self, response, *a, **kw):
        """
        Perform a SOAP request with a canned response.
        """
        return perform_soap_request(
            http_request_full=partial(
                self._http_request_full, response), *a, **kw)

    def test_success(self):
        """
        `perform_soap_request` issues a SOAP request, over HTTP, to a URI, sets
        the ``SOAPAction`` header and parses the response as a SOAP envelope.
        """
        response = MockResponse.build(http.OK, 'response', 'response_header')
        body, header = self.successResultOf(
            self._perform_soap_request(response, 'uri', 'action', 'request'))
        self.assertEqual([
            ('uri',
             tostring(soap_envelope('request')),
             {'SOAPAction': 'action',
              'Content-Type': 'text/xml; charset="utf-8"'})],
            self.requests)
        self.assertEqual(SOAP_ENV.Body.text, body.tag)
        self.assertEqual('response', body.text)
        self.assertEqual(SOAP_ENV.Header.text, header.tag)
        self.assertEqual('response_header', header.text)

    def test_response_not_xml(self):
        """
        `perform_soap_request` raises `xml.etree.ElementTree.ParseError` if the
        response is not valid XML.
        """
        response = MockResponse(http.OK, 'hello')
        self.failureResultOf(
            self._perform_soap_request(response, 'uri', 'action', 'request'),
            ParseError)

    def test_response_no_body(self):
        """
        `perform_soap_request` raises `SoapFault` if the response contains no
        SOAP body element..
        """
        response = MockResponse(http.OK, tostring(SOAP_ENV.Envelope('hello')))
        f = self.failureResultOf(
            self._perform_soap_request(response, 'uri', 'action', 'request'),
            SoapFault)
        self.assertEqual('soapenv:Client', f.value.code)
        self.assertEqual('Malformed SOAP request', f.getErrorMessage())

    def test_fault(self):
        """
        `perform_soap_request` raises `SoapFault`, parsed from the ``Fault``
        element in the response, if the response HTTP status is ``500 Internal
        server error``.
        """
        response = MockResponse.build(
            http.INTERNAL_SERVER_ERROR, soap_fault('soapenv:Server', 'Whoops'))
        f = self.failureResultOf(
            self._perform_soap_request(response, 'uri', 'action', 'request'),
            SoapFault)
        self.assertEqual(
            ('soapenv:Server', 'Whoops'),
            (f.value.code, f.getErrorMessage()))

    def test_expected_fault(self):
        """
        `perform_soap_request` raises a `SoapFault` subclass when a SOAP fault
        detail matches one of the expected fault types.
        """
        detail = L.ToyFaultDetail(
                L.foo('a'),
                L.bar('b'))
        response = MockResponse.build(
            http.INTERNAL_SERVER_ERROR,
            soap_fault('soapenv:Server', 'Whoops', detail=detail))
        f = self.failureResultOf(
            self._perform_soap_request(
                response, 'uri', 'action', 'request',
                expected_faults=[TestFault]),
            TestFault)
        self.assertEqual(
            ('soapenv:Server', 'Whoops'),
            (f.value.code, f.getErrorMessage()))
        parsed_detail = f.value.parsed_detail
        self.assertEqual(
            ('a', 'b'),
            (parsed_detail.foo, parsed_detail.bar))



import os

from vumi.servicemaker import (
    VumiOptions, StartWorkerOptions, VumiWorkerServiceMaker)
from vumi import servicemaker
from vumi.tests.helpers import VumiTestCase


class OptionsTestCase(VumiTestCase):
    "Base class for handling options files"

    def mk_config_file(self, name, lines=None):
        self.config_file[name] = self.mktemp()
        tempfile = open(self.config_file[name], 'w')
        if lines is not None:
            tempfile.write('\n'.join(lines))
        tempfile.close()

    def setUp(self):
        self.config_file = {}


class TestVumiOptions(OptionsTestCase):
    def test_defaults(self):
        options = VumiOptions()
        options.parseOptions([])
        self.assertEqual({}, options.opts)
        self.assertEqual(VumiOptions.default_vumi_options,
                         options.vumi_options)

    def test_override(self):
        options = VumiOptions()
        options.parseOptions(['--hostname', 'blah',
                              '--username', 'haxor'])
        self.assertEqual({}, options.opts)
        self.assertEqual(dict(VumiOptions.default_vumi_options,
                              username='haxor', hostname='blah'),
                         options.vumi_options)

    def test_config_file(self):
        options = VumiOptions()
        self.mk_config_file('vumi', ["username: foo", "password: bar"])
        options.parseOptions(['--vumi-config', self.config_file['vumi']])
        self.assertEqual({}, options.opts)
        self.assertEqual(dict(VumiOptions.default_vumi_options,
                              username='foo', password='bar'),
                         options.vumi_options)

    def test_config_file_override(self):
        self.mk_config_file('vumi', ["username: foo", "password: bar"])
        options = VumiOptions()
        options.parseOptions(['--vumi-config', self.config_file['vumi'],
                              '--hostname', 'blah',
                              '--username', 'haxor'])
        self.assertEqual({}, options.opts)
        self.assertEqual(dict(VumiOptions.default_vumi_options,
                              username='haxor', password='bar',
                              hostname='blah'),
                         options.vumi_options)


class OldConfigWorker(object):
    """Dummy worker for testing --worker-help on old workers
    without CONFIG_CLASS"""


class DummyConfigClass(object):
    """Extra bit of doc string for testing --worker-help."""


class NewConfigWorker(object):
    """Dummy worker for testing --worker-help on new workers
    that support CONFIG_CLASS"""

    CONFIG_CLASS = DummyConfigClass


class TestStartWorkerOptions(OptionsTestCase):
    def test_override(self):
        options = StartWorkerOptions()
        options.parseOptions(['--worker-class', 'foo.FooWorker',
                              '--set-option', 'blah:bleh',
                              '--set-option', 'hungry:supper',
                              ])
        self.assertEqual(VumiOptions.default_vumi_options,
                         options.vumi_options)
        self.assertEqual({}, options.opts)
        self.assertEqual({
                'blah': 'bleh',
                'hungry': 'supper',
                }, options.worker_config)

    def test_config_file(self):
        self.mk_config_file('worker',
                            ["transport_name: sphex", "blah: thingy"])
        options = StartWorkerOptions()
        options.parseOptions(['--worker-class', 'foo.FooWorker',
                              '--config', self.config_file['worker'],
                              ])
        self.assertEqual(VumiOptions.default_vumi_options,
                         options.vumi_options)
        self.assertEqual({}, options.opts)
        self.assertEqual({
                'transport_name': 'sphex',
                'blah': 'thingy',
                }, options.worker_config)

    def test_config_file_with_include(self):
        # Create include file
        self.mk_config_file('provider_prefixes', [
            '"+27606": "vodacom"',
            '"+27603": "mtn"'])
        # Create root file
        self.mk_config_file('worker', [
            'transport_name: foo',
            'middleware:',
            '  - provider_inbound_mw: vumi.provider.foo',
            'provider_inbound_mw:'])
        # Find rel path between include and root files
        rel_path_providers = os.path.relpath(
            self.config_file['provider_prefixes'],
            os.path.dirname(self.config_file['worker']))
        # Add rel path to root file
        with open(self.config_file['worker'], 'a') as f:
            f.write('    !include %s\n' % rel_path_providers)
        # Process config files
        options = StartWorkerOptions()
        options.parseOptions([
            '--worker-class', 'foo.fooWorker',
            '--config', self.config_file['worker']])
        self.assertEqual(
            options.worker_config['provider_inbound_mw'],
            {
                '+27606': 'vodacom',
                '+27603': 'mtn',
            })

    def test_config_file_override(self):
        self.mk_config_file('worker',
                            ["transport_name: sphex", "blah: thingy"])
        options = StartWorkerOptions()
        options.parseOptions(['--worker-class', 'foo.FooWorker',
                              '--config', self.config_file['worker'],
                              '--set-option', 'blah:bleh',
                              '--set-option', 'hungry:supper',
                              ])
        self.assertEqual(VumiOptions.default_vumi_options,
                         options.vumi_options)
        self.assertEqual({}, options.opts)
        self.assertEqual({
                'transport_name': 'sphex',
                'blah': 'bleh',
                'hungry': 'supper',
                }, options.worker_config)

    def test_with_vumi_opts(self):
        self.mk_config_file('vumi', ["username: foo", "password: bar"])
        self.mk_config_file('worker', ["transport_name: sphex"])
        options = StartWorkerOptions()
        options.parseOptions(['--vumi-config', self.config_file['vumi'],
                              '--hostname', 'blah',
                              '--username', 'haxor',
                              '--worker-class', 'foo.FooWorker',
                              '--config', self.config_file['worker'],
                              ])
        self.assertEqual(dict(VumiOptions.default_vumi_options,
                              username='haxor', password='bar',
                              hostname='blah'),
                         options.vumi_options)
        self.assertEqual({}, options.opts)

    def check_worker_help(self, worker_class_name, expected_emits):
        exits, emits = [], []
        options = StartWorkerOptions()
        options.exit = lambda: exits.append(None)  # stub out exit
        options.emit = lambda text: emits.append(text)
        options.parseOptions(['--worker-class', worker_class_name,
                              '--worker-help',
                              ])
        self.assertEqual(len(exits), 1)
        self.assertEqual(emits, expected_emits)

    def test_old_style_config_worker_help(self):
        self.check_worker_help('vumi.tests.test_servicemaker.OldConfigWorker',
                               [OldConfigWorker.__doc__, ""])

    def test_new_style_config_worker_help(self):
        self.check_worker_help('vumi.tests.test_servicemaker.NewConfigWorker',
                               [NewConfigWorker.__doc__,
                                NewConfigWorker.CONFIG_CLASS.__doc__,
                                ""])


class DummyService(object):
    name = "Dummy"


class TestVumiWorkerServiceMaker(OptionsTestCase):

    def test_make_worker(self):
        self.mk_config_file('worker', ["transport_name: sphex"])
        options = StartWorkerOptions()
        options.parseOptions(['--worker-class', 'vumi.demos.words.EchoWorker',
                              '--config', self.config_file['worker'],
                              ])
        maker = VumiWorkerServiceMaker()
        worker = maker.makeService(options)
        self.assertEqual({'transport_name': 'sphex'}, worker.config)

    def test_make_worker_with_sentry(self):
        services = []
        dummy_service = DummyService()

        def service(*a, **kw):
            services.append((a, kw))
            return dummy_service

        self.patch(servicemaker, 'SentryLoggerService', service)
        self.mk_config_file('worker', ["transport_name: sphex"])
        options = StartWorkerOptions()
        options.parseOptions(['--worker-class', 'vumi.demos.words.EchoWorker',
                              '--config', self.config_file['worker'],
                              '--sentry', 'http://1:2@example.com/2/',
                              ])
        maker = VumiWorkerServiceMaker()
        worker = maker.makeService(options)
        self.assertEqual(services, [
                (('http://1:2@example.com/2/',
                  'echoworker',
                  'global:echoworker'), {})
        ])
        self.assertTrue(dummy_service in worker.services)



"""Tests for vumi.persist.session."""

import time

from twisted.internet.defer import inlineCallbacks

from vumi.components.session import SessionManager
from vumi.tests.helpers import VumiTestCase, PersistenceHelper


class TestSessionManager(VumiTestCase):
    @inlineCallbacks
    def setUp(self):
        self.persistence_helper = self.add_helper(PersistenceHelper())
        self.manager = yield self.persistence_helper.get_redis_manager()
        yield self.manager._purge_all()  # Just in case
        self.sm = SessionManager(self.manager)
        self.add_cleanup(self.sm.stop)

    @inlineCallbacks
    def test_active_sessions(self):
        def get_sessions():
            return self.sm.active_sessions().addCallback(lambda s: sorted(s))

        def ids():
            return get_sessions().addCallback(lambda s: [x[0] for x in s])

        self.assertEqual((yield ids()), [])
        yield self.sm.create_session("u1")
        self.assertEqual((yield ids()), ["u1"])
         # 10 seconds later
        yield self.sm.create_session("u2", created_at=time.time() + 10)
        self.assertEqual((yield ids()), ["u1", "u2"])

        s1, s2 = yield get_sessions()
        self.assertTrue(s1[1]['created_at'] < s2[1]['created_at'])

    @inlineCallbacks
    def test_schedule_session_expiry(self):
        self.sm.max_session_length = 60.0
        yield self.sm.create_session("u1")

    @inlineCallbacks
    def test_create_and_retrieve_session(self):
        session = yield self.sm.create_session("u1")
        self.assertEqual(sorted(session.keys()), ['created_at'])
        self.assertTrue(time.time() - float(session['created_at']) < 10.0)
        loaded = yield self.sm.load_session("u1")
        self.assertEqual(loaded, session)

    @inlineCallbacks
    def test_create_clears_existing_session(self):
        session = yield self.sm.create_session("u1", foo="bar")
        self.assertEqual(sorted(session.keys()), ['created_at', 'foo'])
        loaded = yield self.sm.load_session("u1")
        self.assertEqual(loaded, session)

        session = yield self.sm.create_session("u1", bar="baz")
        self.assertEqual(sorted(session.keys()), ['bar', 'created_at'])
        loaded = yield self.sm.load_session("u1")
        self.assertEqual(loaded, session)

    @inlineCallbacks
    def test_save_session(self):
        test_session = {"foo": 5, "bar": "baz"}
        yield self.sm.create_session("u1")
        yield self.sm.save_session("u1", test_session)
        session = yield self.sm.load_session("u1")
        self.assertTrue(session.pop('created_at') is not None)
        # Redis saves & returns all session values as strings
        self.assertEqual(session, dict([map(str, kvs) for kvs
                                        in test_session.items()]))



# -*- test-case-name: vumi.application.tests.test_sandbox_rlimiter -*-

"""NOTE:

    This module is also used as a standalone Python program that is executed by
    the sandbox machinery. It must never, ever import non-stdlib modules.
"""

import os
import sys
import json
import signal
import resource


class SandboxRlimiter(object):
    """This reads rlimits in from stdin, applies them and then execs a
    new executable.

    It's necessary because Twisted's spawnProcess has no equivalent of
    the `preexec_fn` argument to :class:`subprocess.POpen`.

    See http://twistedmatrix.com/trac/ticket/4159.
    """
    def __init__(self, argv, env):
        start = argv.index('--') + 1
        self._executable = argv[start]
        self._args = [self._executable] + argv[start + 1:]
        self._env = env

    def _apply_rlimits(self):
        data = os.environ[self._SANDBOX_RLIMITS_]
        rlimits = json.loads(data) if data.strip() else {}
        for rlimit, (soft, hard) in rlimits.iteritems():
            # Cap our rlimits to the maximum allowed.
            rsoft, rhard = resource.getrlimit(int(rlimit))
            soft = min(soft, rsoft)
            hard = min(hard, rhard)
            resource.setrlimit(int(rlimit), (soft, hard))

    def _reset_signals(self):
        # reset all signal handlers to their defaults
        for i in range(1, signal.NSIG):
            if signal.getsignal(i) == signal.SIG_IGN:
                signal.signal(i, signal.SIG_DFL)

    def _sanitize_fds(self):
        # close everything except stdin, stdout and stderr
        maxfds = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        os.closerange(3, maxfds)

    def execute(self):
        self._apply_rlimits()
        self._restore_child_env(os.environ)
        self._sanitize_fds()
        self._reset_signals()
        os.execvpe(self._executable, self._args, self._env)

    _SANDBOX_RLIMITS_ = "_SANDBOX_RLIMITS_"

    @classmethod
    def _override_child_env(cls, env, rlimits):
        """Put RLIMIT config in the env."""

        env[cls._SANDBOX_RLIMITS_] = json.dumps(rlimits)

    @classmethod
    def _restore_child_env(cls, env):
        """Remove RLIMIT config."""
        del env[cls._SANDBOX_RLIMITS_]

    @classmethod
    def script_name(cls):
        # we need to pass Python the actual filename of this script
        # (rather than using -m __name__) so that is doesn't import
        # Twisted's reactor (since that causes errors when we close
        # all the file handles if using certain reactors).
        script_name = __file__
        if script_name.endswith('.pyc') or script_name.endswith('.pyo'):
            script_name = script_name[:-len('.pyc')] + '.py'
        return script_name

    @classmethod
    def spawn(cls, reactor, protocol, executable, rlimits, **kwargs):
        # spawns a SandboxRlimiter, connectionMade then passes the rlimits
        # through to stdin and the SandboxRlimiter applies them
        args = kwargs.pop('args', [])
        # the -u for unbuffered I/O is important (otherwise the process
        # execed will be very confused about where its stdin data has
        # gone)
        args = [sys.executable, '-u', cls.script_name(), '--'] + args
        env = kwargs.pop('env', {})
        cls._override_child_env(env, rlimits)
        reactor.spawnProcess(protocol, sys.executable, args=args, env=env,
                             **kwargs)


if __name__ == "__main__":
    rlimiter = SandboxRlimiter(sys.argv, os.environ)
    rlimiter.execute()



# -*- test-case-name: vumi.scripts.tests.test_db_backup -*-
import sys
import json
import pkg_resources
import traceback
import re
import time
import calendar
import copy
from datetime import datetime

import yaml
from twisted.python import usage

from vumi.persist.redis_manager import RedisManager
from vumi.errors import ConfigError


def vumi_version():
    vumi = pkg_resources.get_distribution("vumi")
    return str(vumi)


class KeyHandler(object):

    REDIS_TYPES = ('string', 'list', 'set', 'zset', 'hash')

    def __init__(self):
        self._get_handlers = dict((ktype, getattr(self, '%s_get' % ktype))
                                  for ktype in self.REDIS_TYPES)
        self._set_handlers = dict((ktype, getattr(self, '%s_set' % ktype))
                                  for ktype in self.REDIS_TYPES)

    def dump_key(self, redis, key):
        key_type = redis.type(key)
        record = {
            'type': key_type,
            'key': key,
            'value': self._get_handlers[key_type](redis, key),
            'ttl': redis.ttl(key),
        }
        return record

    def restore_key(self, redis, record, ttl_offset=0):
        key, key_type, ttl = record['key'], record['type'], record['ttl']
        if ttl is not None:
            ttl -= ttl_offset
            if ttl <= 0:
                return
        self._set_handlers[key_type](redis, key, record['value'])
        if ttl is not None:
            redis.expire(key, int(round(ttl)))

    def record_okay(self, record):
        if not isinstance(record, dict):
            return False
        for key in ('type', 'key', 'value', 'ttl'):
            if key not in record:
                return False
        return True

    def string_get(self, redis, key):
        return redis.get(key)

    def string_set(self, redis, key, value):
        redis.set(key, value)

    def list_get(self, redis, key):
        return redis.lrange(key, 0, -1)

    def list_set(self, redis, key, value):
        for item in value:
            redis.rpush(key, item)

    def set_get(self, redis, key):
        return sorted(redis.smembers(key))

    def set_set(self, redis, key, value):
        for item in value:
            redis.sadd(key, item)

    def zset_get(self, redis, key):
        return redis.zrange(key, 0, -1, withscores=True)

    def zset_set(self, redis, key, value):
        for item, score in value:
            redis.zadd(key, **{item.encode('utf8'): score})

    def hash_get(self, redis, key):
        return redis.hgetall(key)

    def hash_set(self, redis, key, value):
        redis.hmset(key, value)


class BackupDbsCmd(usage.Options):

    synopsis = "<db-config.yaml> <db-backup-output.json>"

    optFlags = [
        ["not-sorted", None, "Don't sort keys when doing backup."],
    ]

    def parseArgs(self, db_config, db_backup):
        self.db_config = yaml.safe_load(open(db_config))
        self.db_backup = open(db_backup, "wb")
        self.redis_config = self.db_config.get('redis_manager', {})

    def header(self, cfg):
        return {
            'vumi_version': vumi_version(),
            'format': 'LF separated JSON',
            'backup_type': 'redis',
            'timestamp': cfg.get_utcnow().isoformat(),
            'sorted': not bool(self['not-sorted']),
            'redis_config': self.redis_config,
        }

    def write_line(self, data):
        self.db_backup.write(json.dumps(data))
        self.db_backup.write("\n")

    def run(self, cfg):
        cfg.emit("Backing up dbs ...")
        redis = cfg.get_redis(self.redis_config)
        key_handler = KeyHandler()
        keys = redis.keys()
        if not self.opts['not-sorted']:
            keys = sorted(keys)
        self.write_line(self.header(cfg))
        for key in keys:
            record = key_handler.dump_key(redis, key)
            self.write_line(record)
        self.db_backup.close()
        cfg.emit("Backed up %d keys." % (len(keys),))


class RestoreDbsCmd(usage.Options):

    synopsis = "<db-config.yaml> <db-backup.json>"

    optFlags = [
        ["purge", None, "Purge all keys from the redis manager before "
                        "restoring."],
        ["frozen-ttls", None, "Restore TTLs of keys to the same value they "
                              "had when the backup was created, disregarding "
                              "how much time has passed since the backup was "
                              "created. The default is adjust TTLs by the "
                              "amount of time that has passed and to expire "
                              "keys whose TTLs are then zero or negative."],
    ]

    def parseArgs(self, db_config, db_backup):
        self.db_config = yaml.safe_load(open(db_config))
        self.db_backup = open(db_backup, "rb")
        self.redis_config = self.db_config.get('redis_manager', {})

    def check_header(self, header):
        if header is None:
            return None, "Header not found."
        try:
            header = json.loads(header)
        except Exception:
            return None, "Header not JSON."
        if not isinstance(header, dict):
            return None, "Header not JSON dict."
        if 'backup_type' not in header:
            return None, "Header missing backup_type."
        if header['backup_type'] != 'redis':
            return None, "Only redis backup type currently supported."
        return header, None

    def seconds_from_now(self, iso_timestamp):
        seconds_timestamp, _dot, _milliseconds = iso_timestamp.partition('.')
        time_of_backup = time.strptime(seconds_timestamp, "%Y-%m-%dT%H:%M:%S")
        return time.time() - calendar.timegm(time_of_backup)

    def run(self, cfg):
        line_iter = iter(self.db_backup)
        try:
            header = line_iter.next()
        except StopIteration:
            header = None

        header, error = self.check_header(header)
        if error is not None:
            cfg.emit(error)
            cfg.emit("Aborting restore.")
            return

        if self.opts['frozen-ttls']:
            ttl_offset = 0
        else:
            ttl_offset = self.seconds_from_now(header['timestamp'])

        cfg.emit("Restoring dbs ...")
        redis = cfg.get_redis(self.redis_config)
        if self.opts['purge']:
            redis._purge_all()
        key_handler = KeyHandler()
        keys, skipped = 0, 0
        for i, line in enumerate(line_iter):
            try:
                record = json.loads(line)
            except Exception:
                excinfo = sys.exc_info()
                for s in traceback.format_exception(*excinfo):
                    cfg.emit(s)
                skipped += 1
                continue
            if not key_handler.record_okay(record):
                cfg.emit("Skipping bad backup record on line %d." % (i + 1,))
                skipped += 1
                continue
            key_handler.restore_key(redis, record, ttl_offset)
            keys += 1

        cfg.emit("%d keys successfully restored." % keys)
        if skipped != 0:
            cfg.emit("WARNING: %d bad backup lines skipped." % skipped)


class MigrateDbsCmd(usage.Options):

    synopsis = ("<migration-config.yaml> <db-backup.json>"
                " <migrated-backup.json>")

    def parseArgs(self, migration_config, db_backup, migrated_backup):
        self.migration_config = yaml.safe_load(open(migration_config))
        self.db_backup = open(db_backup, "rb")
        self.migrated_backup = open(migrated_backup, "wb")

    def postOptions(self):
        self.rules = self.create_rules(self.migration_config)

    def make_rule_drop(self, kw):
        key_regex = re.compile(kw['key'])

        def rule(record):
            if key_regex.match(record['key']):
                return True, None
            return False, record

        return rule

    def make_rule_rename(self, kw):
        from_regex = re.compile(kw['from'])
        to_template = kw['to']

        def rule(record):
            key = record['key']
            record['key'] = from_regex.sub(to_template, key)
            if record['key'] == key:
                return False, record
            return True, record

        return rule

    def create_rules(self, migration_config):
        rules = []
        for rule in migration_config['rules']:
            kw = rule.copy()
            rule_name = kw.pop('type')
            rule_maker = getattr(self, "make_rule_%s" % rule_name, None)
            if rule_maker is None:
                raise ConfigError("Unknown rule type %r" % rule_name)
            rules.append(rule_maker(kw))
        return rules

    def apply_rules(self, record):
        for rule in self.rules:
            done, record = rule(record)
            if done:
                break
        return record

    def write_line(self, record):
        self.migrated_backup.write(json.dumps(record))
        self.migrated_backup.write("\n")

    def run(self, cfg):
        line_iter = iter(self.db_backup)
        try:
            header = line_iter.next()
        except StopIteration:
            cfg.emit("No header in backup.")
            return
        self.write_line(json.loads(header))

        cfg.emit("Migrating backup ...")
        changed, processed = 0, 0
        for data in line_iter:
            record = json.loads(data)
            new_record = self.apply_rules(copy.deepcopy(record))
            if record != new_record:
                changed += 1
            processed += 1
            if new_record is not None:
                self.write_line(new_record)
        self.migrated_backup.close()

        cfg.emit("Summary of changes:")
        cfg.emit("  %d records processed." % processed)
        cfg.emit("  %d records altered." % changed)


class PrefixTree(object):
    def __init__(self):
        self._root = {}

    def add_key(self, key):
        node = self._root
        for c in key:
            if c not in node:
                node[c] = {}
            node = node[c]

    def compress_edges(self, edge_pattern):
        edge_regex = re.compile(edge_pattern)
        new_root = {}

        stack = [(self._root, new_root, '')]
        while stack:
            current_old, current_new, prefix = stack.pop()
            for c, next_old in current_old.iteritems():
                edge = prefix + c
                if edge_regex.match(edge):
                    current_new[edge] = next_new = {}
                    stack.append((next_old, next_new, ''))
                elif next_old:
                    stack.append((next_old, current_new, edge))
                else:
                    current_new[edge] = {}

        self._root = new_root

    def _print_tree(self, emit, indent, node, level):
        full_indent = indent * level
        for edge, node_edge in sorted(node.items()):
            sub_trees = dict((k, v) for k, v in node_edge.items() if v)
            leaves = len(node_edge) - len(sub_trees)
            emit("%s%s%s" % (full_indent, edge,
                             " (%d leaves)" % leaves if leaves else ""))
            if sub_trees:
                self._print_tree(emit, indent, sub_trees, level + 1)

    def print_tree(self, emit, indent="  "):
        return self._print_tree(emit, indent, self._root, 0)


class AnalyzeCmd(usage.Options):

    synopsis = "<db-backup-output.json>"

    optParameters = [
        ["separators", "s", "[:#]",
         "Regular expression for allowed key part separators."],
    ]

    def parseArgs(self, db_backup):
        self.db_backup = open(db_backup, "rb")

    def run(self, cfg):
        backup_lines = iter(self.db_backup)
        try:
            backup_lines.next()  # skip header
        except StopIteration:
            cfg.emit("No header found. Aborting.")
            return

        tree = PrefixTree()
        for i, line in enumerate(backup_lines):
            try:
                key = json.loads(line)['key']
            except:
                cfg.emit("Bad record %d: %r" % (i, line))
                continue
            tree.add_key(key)
        edge_pattern = r".*%s" % self.opts['separators']
        tree.compress_edges(edge_pattern)

        cfg.emit("Keys:")
        cfg.emit("-----")
        tree.print_tree(cfg.emit)


class Options(usage.Options):
    subCommands = [
        ["backup", None, BackupDbsCmd,
         "Backup databases."],
        ["restore", None, RestoreDbsCmd,
         "Restore databases."],
        ["migrate", None, MigrateDbsCmd,
         "Rename keys in a database backup."],
        ["analyze", None, AnalyzeCmd,
         "Analyze a database backup."],
    ]

    longdesc = """Back-up and restore utility for Vumi
                  Redis (and maybe later Riak) data stores."""

    def postOptions(self):
        if self.subCommand is None:
            raise usage.UsageError("Please specify a sub-command.")


class ConfigHolder(object):
    def __init__(self, options):
        self.options = options

    def emit(self, s):
        print s

    def get_utcnow(self):
        return datetime.utcnow()

    def get_redis(self, config):
        return RedisManager.from_config(config)

    def run(self):
        self.options.subOptions.run(self)


if __name__ == '__main__':
    try:
        options = Options()
        options.parseOptions()
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        sys.exit(1)

    cfg = ConfigHolder(options)
    cfg.run()



# -*- test-case-name: vumi.persist.tests.test_redis_manager -*-

import redis

from vumi.persist.redis_base import Manager
from vumi.persist.fake_redis import FakeRedis
from vumi.utils import flatten_generator


class VumiRedis(redis.Redis):
    """
    Custom Vumi redis client implementation.
    """

    def setex(self, key, seconds, value):
        """
        The underlying .setex() signature doesn't match our implementation
        in the txredis manager. This wrapper swaps the last two parameters,
        seconds and value, so that they do.
        """
        return super(VumiRedis, self).setex(key, value, seconds)

    def scan(self, cursor, match=None, count=None):
        """
        Scan through all the keys in the database returning those that
        match the pattern ``match``. The ``cursor`` specifies where to
        start a scan and ``count`` determines how much work to do looking
        for keys on each scan. ``cursor`` may be ``None`` or ``'0'`` to
        indicate a new scan. Any other value should be treated as an opaque
        string.

        .. note::

           Requires redis server 2.8 or later.
        """
        args = []
        if cursor is None:
            cursor = '0'
        if match is not None:
            args.extend(("MATCH", match))
        if count is not None:
            args.extend(("COUNT", count))
        cursor, keys = self.execute_command("SCAN", cursor, *args)
        if cursor == '0' or cursor == 0:
            cursor = None
        return (cursor, keys)


class RedisManager(Manager):

    call_decorator = staticmethod(flatten_generator)

    @classmethod
    def _fake_manager(cls, fake_redis, manager_config):
        if fake_redis is None:
            fake_redis = FakeRedis(async=False)
        manager_config['config']['FAKE_REDIS'] = fake_redis
        manager = cls(fake_redis, **manager_config)
        # Because ._close() assumes a real connection.
        manager._close = fake_redis.teardown
        return manager

    @classmethod
    def _manager_from_config(cls, config, manager_config):
        """Construct a manager from a dictionary of options.

        :param dict config:
            Dictionary of options for the manager.
        :param str key_prefix:
            Key prefix for namespacing.
        """

        return cls(VumiRedis(**config), **manager_config)

    def _close(self):
        """Close redis connection."""
        # Close all the connections this client may have open.
        self._client.connection_pool.disconnect()

    def _purge_all(self):
        """Delete *ALL* keys whose names start with this manager's key prefix.

        Use only in tests.
        """
        for key in self.keys():
            self.delete(key)

    def _make_redis_call(self, call, *args, **kw):
        """Make a redis API call using the underlying client library.
        """
        return getattr(self._client, call)(*args, **kw)

    def _filter_redis_results(self, func, results):
        """Filter results of a redis call.
        """
        return func(results)
